from typing import Self, override
from dataclasses import asdict

from luma.core.super import Estimator, Evaluator, Optimizer, Supervised
from luma.interface.typing import Matrix, Tensor, Vector
from luma.interface.util import InitUtil
from luma.metric.classification import Accuracy

from luma.neural import loss
from luma.neural.base import Loss, NeuralModel
from luma.neural.block import ConvBlock2D, DenseBlock, ConvBlockArgs, DenseBlockArgs
from luma.neural.layer import (
    Activation,
    Dense,
    Flatten,
    LocalResponseNorm,
    Sequential,
)


__all__ = ("_AlexNet", "_ZFNet")


class _AlexNet(Estimator, Supervised, NeuralModel):
    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.ReLU,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 1000,
        batch_size: int = 128,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0.0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int = None,
        deep_verbose: bool = False,
    ) -> None:
        self.activation = activation
        self.optimizer = optimizer
        self.loss = loss
        self.initializer = initializer
        self.out_features = out_features
        self.lambda_ = lambda_
        self.dropout_rate = dropout_rate
        self.shuffle = shuffle
        self.random_state = random_state
        self._fitted = False

        super().__init__(
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            early_stopping,
            patience,
            deep_verbose,
        )
        super().__init_model__()
        self.model = Sequential()
        self.optimizer.set_params(learning_rate=self.learning_rate)
        self.model.set_optimizer(optimizer=self.optimizer)

        self.feature_sizes_ = [
            [3, 96, 256, 384, 384, 256],
            [256 * 6 * 6, 4096, 4096, self.out_features],
        ]
        self.feature_shapes_ = [
            self._get_feature_shapes(self.feature_sizes_[0]),
            self._get_feature_shapes(self.feature_sizes_[1]),
        ]

        self.set_param_ranges(
            {
                "out_features": ("0<,+inf", int),
                "batch_size": ("0<,+inf", int),
                "n_epochs": ("0<,+inf", int),
                "learning_rate": ("0<,+inf", None),
                "valid_size": ("0<,<1", None),
                "dropout_rate": ("0,1", None),
                "lambda_": ("0,+inf", None),
                "patience": (f"0<,+inf", int),
            }
        )
        self.check_param_ranges()
        self._build_model()

    def _build_model(self) -> None:
        conv_3x3_no_pool_arg = ConvBlockArgs(
            filter_size=3,
            stride=1,
            activation=self.activation,
            initializer=self.initializer,
            padding="same",
            lambda_=self.lambda_,
            do_batch_norm=False,
            do_pooling=False,
            random_state=self.random_state,
        )
        dense_args = DenseBlockArgs(
            activation=self.activation,
            lambda_=self.lambda_,
            do_batch_norm=False,
            dropout_rate=self.dropout_rate,
            random_state=self.random_state,
        )

        self.model += (
            "ConvBlock_1",
            ConvBlock2D(
                3,
                96,
                filter_size=11,
                stride=4,
                activation=self.activation,
                initializer=self.initializer,
                padding="valid",
                lambda_=self.lambda_,
                do_batch_norm=False,
                pool_filter_size=3,
                pool_stride=2,
                pool_mode="max",
                random_state=self.random_state,
            ),
        )
        self.model += (
            "LRN_1",
            LocalResponseNorm(depth=5),
        )

        self.model += (
            "ConvBlock_2",
            ConvBlock2D(
                96,
                256,
                filter_size=5,
                stride=1,
                activation=self.activation,
                initializer=self.initializer,
                padding="same",
                lambda_=self.lambda_,
                do_batch_norm=False,
                pool_filter_size=3,
                pool_stride=2,
                pool_mode="max",
                random_state=self.random_state,
            ),
        )
        self.model += (
            "LRN_2",
            LocalResponseNorm(depth=5),
        )

        self.model += (
            "ConvBlock_3",
            ConvBlock2D(256, 384, **asdict(conv_3x3_no_pool_arg)),
        )
        self.model += (
            "LRN_3",
            LocalResponseNorm(depth=5),
        )

        self.model += (
            "ConvBlock_4",
            ConvBlock2D(384, 384, **asdict(conv_3x3_no_pool_arg)),
        )
        self.model += (
            "LRN_4",
            LocalResponseNorm(depth=5),
        )

        self.model += (
            "ConvBlock_5",
            ConvBlock2D(
                384,
                256,
                filter_size=3,
                stride=1,
                activation=self.activation,
                initializer=self.initializer,
                padding="same",
                lambda_=self.lambda_,
                do_batch_norm=False,
                pool_filter_size=3,
                pool_stride=2,
                pool_mode="max",
                random_state=self.random_state,
            ),
        )
        self.model += (
            "LRN_5",
            LocalResponseNorm(depth=5),
        )

        self.model += Flatten()
        self.model += (
            "DenseBlock_1",
            DenseBlock(256 * 6 * 6, 4096, **asdict(dense_args)),
        )
        self.model += (
            "DenseBlock_2",
            DenseBlock(4096, 4096, **asdict(dense_args)),
        )
        self.model += Dense(
            4096,
            self.out_features,
            lambda_=self.lambda_,
            random_state=self.random_state,
        )

    @Tensor.force_dim(4)
    def fit(self, X: Tensor, y: Matrix) -> Self:
        return super(_AlexNet, self).fit_nn(X, y)

    @override
    @Tensor.force_dim(4)
    def predict(self, X: Tensor, argmax: bool = True) -> Matrix | Vector:
        return super(_AlexNet, self).predict_nn(X, argmax)

    @override
    @Tensor.force_dim(4)
    def score(
        self,
        X: Tensor,
        y: Matrix,
        metric: Evaluator = Accuracy,
        argmax: bool = True,
    ) -> float:
        return super(_AlexNet, self).score_nn(X, y, metric, argmax)


class _ZFNet(Estimator, Supervised, NeuralModel):
    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.ReLU,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 1000,
        batch_size: int = 128,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0.0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int = None,
        deep_verbose: bool = False,
    ) -> None:
        self.activation = activation
        self.optimizer = optimizer
        self.loss = loss
        self.initializer = initializer
        self.out_features = out_features
        self.lambda_ = lambda_
        self.dropout_rate = dropout_rate
        self.shuffle = shuffle
        self.random_state = random_state
        self._fitted = False

        super().__init__(
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            early_stopping,
            patience,
            deep_verbose,
        )
        super().__init_model__()
        self.model = Sequential()
        self.optimizer.set_params(learning_rate=self.learning_rate)
        self.model.set_optimizer(optimizer=self.optimizer)

        self.feature_sizes_ = [
            [3, 96, 256, 384, 384, 256],
            [256 * 6 * 6, 4096, 4096, self.out_features],
        ]
        self.feature_shapes_ = [
            self._get_feature_shapes(self.feature_sizes_[0]),
            self._get_feature_shapes(self.feature_sizes_[1]),
        ]

        self.set_param_ranges(
            {
                "out_features": ("0<,+inf", int),
                "batch_size": ("0<,+inf", int),
                "n_epochs": ("0<,+inf", int),
                "learning_rate": ("0<,+inf", None),
                "valid_size": ("0<,<1", None),
                "dropout_rate": ("0,1", None),
                "lambda_": ("0,+inf", None),
                "patience": (f"0<,+inf", int),
            }
        )
        self.check_param_ranges()
        self._build_model()

    def _build_model(self) -> None:
        conv_3x3_no_pool_arg = ConvBlockArgs(
            filter_size=3,
            stride=1,
            activation=self.activation,
            initializer=self.initializer,
            padding="same",
            lambda_=self.lambda_,
            do_batch_norm=False,
            do_pooling=False,
            random_state=self.random_state,
        )
        dense_args = DenseBlockArgs(
            activation=self.activation,
            lambda_=self.lambda_,
            do_batch_norm=False,
            dropout_rate=self.dropout_rate,
            random_state=self.random_state,
        )

        self.model += (
            "ConvBlock_1",
            ConvBlock2D(
                3,
                96,
                filter_size=7,
                stride=2,
                activation=self.activation,
                initializer=self.initializer,
                padding="valid",
                lambda_=self.lambda_,
                do_batch_norm=False,
                pool_filter_size=3,
                pool_stride=2,
                pool_mode="max",
                random_state=self.random_state,
            ),
        )
        self.model += (
            "LRN_1",
            LocalResponseNorm(depth=5),
        )

        self.model += (
            "ConvBlock_2",
            ConvBlock2D(
                96,
                256,
                filter_size=5,
                stride=2,
                activation=self.activation,
                initializer=self.initializer,
                padding="same",
                lambda_=self.lambda_,
                do_batch_norm=False,
                pool_filter_size=3,
                pool_stride=2,
                pool_mode="max",
                random_state=self.random_state,
            ),
        )
        self.model += (
            "LRN_2",
            LocalResponseNorm(depth=5),
        )

        self.model += (
            "ConvBlock_3",
            ConvBlock2D(256, 384, **asdict(conv_3x3_no_pool_arg)),
        )
        self.model += (
            "LRN_3",
            LocalResponseNorm(depth=5),
        )

        self.model += (
            "ConvBlock_4",
            ConvBlock2D(384, 384, **asdict(conv_3x3_no_pool_arg)),
        )
        self.model += (
            "LRN_4",
            LocalResponseNorm(depth=5),
        )

        self.model += (
            "ConvBlock_5",
            ConvBlock2D(
                384,
                256,
                filter_size=3,
                stride=1,
                activation=self.activation,
                initializer=self.initializer,
                padding="same",
                lambda_=self.lambda_,
                do_batch_norm=False,
                pool_filter_size=3,
                pool_stride=2,
                pool_mode="max",
                random_state=self.random_state,
            ),
        )
        self.model += (
            "LRN_5",
            LocalResponseNorm(depth=5),
        )

        self.model += Flatten()
        self.model += (
            "DenseBlock_1",
            DenseBlock(256 * 6 * 6, 4096, **asdict(dense_args)),
        )
        self.model += (
            "DenseBlock_2",
            DenseBlock(4096, 4096, **asdict(dense_args)),
        )
        self.model += Dense(
            4096,
            self.out_features,
            lambda_=self.lambda_,
            random_state=self.random_state,
        )

    @Tensor.force_dim(4)
    def fit(self, X: Tensor, y: Matrix) -> Self:
        return super(_ZFNet, self).fit_nn(X, y)

    @override
    @Tensor.force_dim(4)
    def predict(self, X: Tensor, argmax: bool = True) -> Matrix | Vector:
        return super(_ZFNet, self).predict_nn(X, argmax)

    @override
    @Tensor.force_dim(4)
    def score(
        self,
        X: Tensor,
        y: Matrix,
        metric: Evaluator = Accuracy,
        argmax: bool = True,
    ) -> float:
        return super(_ZFNet, self).score_nn(X, y, metric, argmax)



from typing import Self
import numpy as np

from luma.interface.typing import Matrix
from luma.interface.util import KernelUtil
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.core.super import Estimator, Evaluator, Supervised
from luma.metric.regression import MeanSquaredError


__all__ = ("SVR", "KernelSVR")


class SVR(Estimator, Supervised):
    """
    Support Vector Regressor (SVR) is a supervised machine learning
    algorithm used for regression tasks. It operates by determining a
    hyperplane that best fits the input data, aiming to minimize the
    difference between the predicted and actual values.

    Parameters
    ----------
    `C` : float, default=1.0
        Regularization parameter
    `epsilon` : float, default=0.1
        Epsilon-tube in the training loss function
    `batch_size` : int, default=100
        Size of a single batch
    `learning_rate` : float, default=0.001
        Step-size of gradient descent update
    `max_iter` : int, default=1000
        Number of iteration

    """

    def __init__(
        self,
        C: float = 1.0,
        epsilon: float = 0.1,
        batch_size: int = 100,
        learning_rate: float = 0.001,
        max_iter: int = 1000,
        verbose: bool = False,
    ) -> None:
        self.C = C
        self.epsilon = epsilon
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.verbose = verbose
        self._fitted = False

        self.set_param_ranges(
            {
                "C": ("0,+inf", None),
                "epsilon": ("0,+inf", None),
                "batch_size": ("0<,+inf", int),
                "learning_rate": ("0<,+inf", None),
                "max_iter": ("0<,+inf", int),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Matrix) -> Self:
        m, n = X.shape
        id = np.arange(m)
        np.random.shuffle(id)

        self.bias = 0.0
        self.weight = np.zeros(n)

        for i in range(self.max_iter):
            for batch_point in range(0, m, self.batch_size):
                gradient_w, gradient_b = 0, 0
                for j in range(batch_point, batch_point + self.batch_size):
                    if j >= m:
                        continue
                    idx = id[j]
                    pred = np.dot(self.weight, X[idx].T) + self.bias
                    loss = pred - y[idx]

                    if np.abs(loss) > self.epsilon:
                        gradient_w += self.C * np.sign(loss) * X[idx]
                        gradient_b += self.C * np.sign(loss)

                self.weight -= self.learning_rate * self.weight
                self.weight -= self.learning_rate * gradient_w
                self.bias -= self.learning_rate * gradient_b

            if self.verbose and i % 100 == 0 and i:
                print(f"[SVR] Finished iteration {i}/{self.max_iter}", end=" ")
                print(f"with weight-norm of {np.linalg.norm(self.weight)}")

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return np.dot(X, self.weight) + self.bias

    def score(
        self, X: Matrix, y: Matrix, metric: Evaluator = MeanSquaredError
    ) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)


class KernelSVR(Estimator, Supervised):
    """
    Kernel Support Vector Regression (KernelSVR) is an extension of the
    Support Vector Regression (SVR) algorithm that enables regression on
    non-linearly related data by mapping it into a higher-dimensional space
    using a kernel function. The kernel function allows the algorithm to
    establish a flexible and non-linear regression model in the transformed space.

    Parameters
    ----------
    `C` : float, default=1.0
        Regularization parameter
    `deg` : int, default=3
        Polynomial Degree for polynomial kernel
    `gamma` : float, default=1.0
        Shape parameter of Gaussian curve for RBF kernel
    `coef` : float, default=1.0
        Coefficient for polynomial, sigmoid kernel
    `learning_rate` : float, default=0.001
        Step-size for gradient descent update
    `max_iter` : int, default=1000
        Number of iteration
    `kernel` : FuncType, default="rbf"
        Type of kernel

    """

    def __init__(
        self,
        C: float = 1.0,
        deg: int = 3,
        gamma: float = 1.0,
        coef: float = 1.0,
        learning_rate: float = 0.001,
        max_iter: int = 1000,
        kernel: KernelUtil.FuncType = "rbf",
        verbose: bool = False,
    ) -> None:
        self.C = C
        self.deg = deg
        self.gamma = gamma
        self.coef = coef
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.kernel = kernel
        self.verbose = verbose
        self._kernel_func = None
        self._fitted = False

        self.set_param_ranges(
            {
                "C": ("0<,+inf", None),
                "deg": ("0<,+inf", int),
                "gamma": ("0<,+inf", None),
                "coef": ("-inf,+inf", None),
                "learning_rate": ("0<,+inf", None),
                "max_iter": ("0<,+inf", int),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Matrix) -> Self:
        m, _ = X.shape
        self._set_kernel_func()
        self.alpha = np.random.random(m)
        self.bias = 0

        y_mul_kernel = np.outer(y, y) * self._kernel_func(X, X)
        for i in range(self.max_iter):
            gradient = np.ones(m) - y_mul_kernel.dot(self.alpha)
            self.alpha += self.learning_rate * gradient
            self.alpha = np.clip(self.alpha, 0, self.C)

            if self.verbose and i % 100 == 0 and i:
                print(f"[KernelSVR] Finished iteration {i}/{self.max_iter}", end=" ")
                print(f"with alpha-norm: {np.linalg.norm(self.alpha)}")

        alpha_idx = np.where((self.alpha > 0) & (self.alpha < self.C))[0]
        bias_list = []
        for idx in alpha_idx:
            _append = y[idx] - (self.alpha * y).dot(self._kernel_func(X, X[idx]))
            bias_list.append(_append)
        self.bias = np.mean(bias_list)

        self._X = X
        self._y = y
        self._fitted = True
        return self

    def _linear_kernel(self, xi: Matrix, xj: Matrix) -> Matrix:
        return xi.dot(xj.T)

    def _poly_kernel(self, xi: Matrix, xj: Matrix) -> Matrix:
        return (self.coef + xi.dot(xj.T)) ** self.deg

    def _rbf_kernel(self, xi: Matrix, xj: Matrix) -> Matrix:
        norm = np.linalg.norm(xi[:, np.newaxis] - xj[np.newaxis, :], axis=2)
        return np.exp(-self.gamma * norm**2)

    def _sigmoid_kernel(self, xi: Matrix, xj: Matrix) -> Matrix:
        return np.tanh(2 * xi.dot(xj.T) + self.coef)

    def _set_kernel_func(self) -> None:
        if self.kernel == "linear":
            self._kernel_func = self._linear_kernel
        elif self.kernel == "poly":
            self._kernel_func = self._poly_kernel
        elif self.kernel == "rbf":
            self._kernel_func = self._rbf_kernel
        elif self.kernel == "sigmoid":
            self._kernel_func = self._sigmoid_kernel
        else:
            raise UnsupportedParameterError(self.kernel)

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return (self.alpha * self._y).dot(self._kernel_func(self._X, X)) + self.bias

    def score(
        self, X: Matrix, y: Matrix, metric: Evaluator = MeanSquaredError
    ) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)



from typing import Any, Callable, Iterable, Literal, Self, Type, TypeGuard
from rich.progress import Progress, BarColumn, TextColumn
import numpy as np

from luma.interface.exception import UnsupportedParameterError, InvalidRangeError
from luma.interface.typing import Matrix, Scalar, Vector
from luma.neural import init


__all__ = (
    "DecisionTreeNode",
    "NearestNeighbors",
    "SilhouetteUtil",
    "DBUtil",
    "KernelUtil",
    "ActivationUtil",
    "InitUtil",
    "Clone",
    "ParamRange",
    "TrainProgress",
)


class DecisionTreeNode:
    """
    Internal class for node used in tree-based models.

    Parameters
    ----------
    `feature_index` : int, optional, default=None
        Feature of node
    `threshold` : float, optional, default=None
        Threshold for split point
    `left` : Self, optional, default=None
        Left-child node
    `right` : Self, optional, default=None
        Right-child node
    `value` : Any, optional, default=None
        Most popular label of leaf node

    """

    def __init__(
        self,
        feature_index: int | None = None,
        threshold: float | None = None,
        left: Self | None = None,
        right: Self | None = None,
        value: Any | None = None,
    ) -> None:
        self.feature_index = feature_index
        self.threshold = threshold
        self.left = left
        self.right = right
        self.value = value

    @property
    def isLeaf(self) -> bool:
        return self.value is not None


class NearestNeighbors:
    """
    Internal class for computing nearest neighbors of given data.

    Parameters
    ----------
    `data` : Matrix
        Data to be handled
    `n_neighbors` : int
        Number of nearest neighbors

    """

    def __init__(self, data: Matrix, n_neighbors: int) -> None:
        self.data = data
        self.n_neighbors = n_neighbors
        self._size = data.shape[0]

    @property
    def index_matrix(self) -> Matrix:
        data = self.data
        dist = np.linalg.norm(data[:, np.newaxis, :] - data, axis=2)
        sorted_indices = np.argsort(dist, axis=1)
        return sorted_indices[:, 1 : self.n_neighbors + 1]

    @property
    def adjacent_matrix(self) -> Matrix:
        indices = self.index_matrix
        adj_mat = np.zeros((self._size, self._size))
        for i in range(self._size):
            adj_mat[i, indices[i]] = 1

        return adj_mat.astype(int)


class SilhouetteUtil:
    """
    Internal class for computing various distances used in
    Silhouette Coefficient calculation.

    Parameters
    ----------
    `idx` : int
        Index of a single data point
    `cluster` : int
        Current cluster number
    `labels` : Vector
        Labels assigned by clustering estimator
    `distances` : Matrix
        Square-form distance matrix of the data

    """

    def __init__(
        self, idx: int, cluster: int, labels: Vector, distances: Matrix
    ) -> None:
        self.idx = idx
        self.cluster = cluster
        self.labels = labels
        self.distances = distances

    @property
    def avg_dist_others(self) -> Matrix:
        others = set(self.labels) - {self.cluster}
        sub_avg = [
            np.mean(
                self.distances[self.idx][self.labels == other],
            )
            for other in others
        ]

        return np.mean(sub_avg)

    @property
    def avg_dist_within(self) -> Matrix | int:
        within_cluster = self.distances[self.idx][self.labels == self.cluster]
        if len(within_cluster) <= 1:
            return 0
        return np.mean([dist for dist in within_cluster if dist != 0])


class DBUtil:
    """
    Internal class for supporting Davies-Bouldin Index (DBI) computation.

    Parameters
    ----------
    `data` : Matrix
        Original data
    `labels` : Vector
        Labels assigned by clustering estimator

    """

    def __init__(self, data: Matrix, labels: Vector) -> None:
        self.data = data
        self.labels = labels

    @property
    def cluster_centroids(self) -> Matrix:
        unique_labels = np.unique(self.labels)
        centroids = [
            self.data[self.labels == label].mean(axis=0) for label in unique_labels
        ]
        return Matrix(centroids)

    @property
    def within_cluster_scatter(self) -> Matrix:
        centroids = self.cluster_centroids
        scatter = np.zeros(len(centroids))

        for i, centroid in enumerate(centroids):
            cluster_points = self.data[self.labels == i]
            diff_sq = (cluster_points - centroid) ** 2
            scatter[i] = np.mean(np.sqrt(np.sum(diff_sq, axis=1)))

        return scatter

    @property
    def separation(self) -> Matrix:
        centroids = self.cluster_centroids
        n_clusters = len(centroids)
        separation = np.zeros((n_clusters, n_clusters))

        for i in range(n_clusters):
            for j in range(n_clusters):
                if i == j:
                    continue
                diff_sq = (centroids[i] - centroids[j]) ** 2
                separation[i, j] = np.sqrt(np.sum(diff_sq))

        return separation


class KernelUtil:
    """
    Internal class for kernel methods(tricks).

    This class facilitates transferring kernel type strings
    into actual specific kernel function.

    Parameters
    ----------
    `kernel` : FuncType
        Type of kernel
    `alpha` : float, default=1.0
        Shape parameter for RBF and sigmoid kernels
     `gamma` : float, default=0.0
        Shape parameter of Gaussian curve for RBF kernel
    `coef` : float, default=1.0
        Coefficient for polynomial and sigmoid kernel
    `deg` : int, default=3
        Polynomial Degree for polynomial kernel

    Example
    -------
    >>> util = KernelUtil(kernel='rbf', **params)
    >>> util.kernel_func
    KernelUtil.rbf_kernel: Callable[[Matrix, Matrix | None], Matrix]

    """

    FuncType = Literal[
        "lin",
        "linear",
        "poly",
        "polynoimal",
        "rbf",
        "gaussian",
        "Gaussian",
        "tanh",
        "sigmoid",
        "lap",
        "laplacian",
    ]

    def __init__(
        self,
        kernel: FuncType,
        alpha: float = 1.0,
        gamma: float = 1.0,
        coef: float = 0.0,
        deg: int = 2,
    ) -> None:
        self.kernel = kernel
        self.alpha = alpha
        self.gamma = gamma
        self.coef = coef
        self.deg = deg

    def linear_kernel(self, Xi: Matrix, Xj: Matrix = None) -> Matrix:
        if Xj is None:
            Xj = Xi.copy()
        return np.dot(Xi, Xj.T)

    def polynomial_kernel(self, Xi: Matrix, Xj: Matrix = None) -> Matrix:
        if Xj is None:
            Xj = Xi.copy()
        return (self.gamma * np.dot(Xi, Xj.T) + self.coef) ** self.deg

    def rbf_kernel(self, Xi: Matrix, Xj: Matrix = None) -> Matrix:
        if Xj is None:
            Xj = Xi.copy()
        _left = np.sum(Xi**2, axis=1).reshape(-1, 1)
        _right = np.sum(Xj**2, axis=1) - 2 * np.dot(Xi, Xj.T)

        return np.exp(-self.gamma * (_left + _right))

    def sigmoid_kernel(self, Xi: Matrix, Xj: Matrix = None) -> Matrix:
        if Xj is None:
            Xj = Xi.copy()
        return np.tanh(self.gamma * np.dot(Xi, Xj.T) + self.coef)

    def laplacian_kernel(self, Xi: Matrix, Xj: Matrix = None) -> Matrix:
        if Xj is None:
            Xj = Xi.copy()
        manhattan_dists = np.sum(np.abs(Xi[:, np.newaxis] - Xj), axis=2)

        return np.exp(-self.gamma * manhattan_dists)

    @property
    def kernel_func(self) -> Callable[[Matrix, Matrix | None], Matrix]:
        if self.kernel in ("linear", "lin"):
            return self.linear_kernel
        elif self.kernel in ("poly", "polynomial"):
            return self.polynomial_kernel
        elif self.kernel in ("rbf", "gaussian", "Gaussian"):
            return self.rbf_kernel
        elif self.kernel in ("sigmoid", "tanh"):
            return self.sigmoid_kernel
        elif self.kernel in ("laplacian", "lap"):
            return self.laplacian_kernel
        else:
            raise UnsupportedParameterError(self.kernel)


class Clone:
    """
    A utility class for cloning LUMA models.

    This class creates a copy of a given LUMA model,
    which can be either an Estimator or a Transformer.
    The clone includes all parameters of the original model.
    Optionally, the trained state of the model can also be copied
    if applicable.

    Parameters
    ----------
    `model` : object, optional, default=None
        The model to be cloned
    `pass_fitted` : bool, default=True
        Whether to copy the fitted state of the original model

    Examples
    --------
    >>> original_model = AnyModel(...)
    >>> cloned_model = Clone(model=original_model, pass_fitted=True).get

    """

    def __init__(
        self,
        model: object | None = None,
        pass_fitted: bool = False,
    ) -> None:
        self.model = model
        self.pass_fitted = pass_fitted

    @property
    def get(self) -> object:
        model_cls = type(self.model)
        new_model = model_cls()

        for param, val in self.model.__dict__.items():
            try:
                new_model.set_params(**{param: val})
            except:
                continue

        if hasattr(self.model, "_fitted") and self.pass_fitted:
            new_model._fitted = self.model._fitted

        return new_model


class ParamRange:
    """
    A utility class for setting and checking the range of a specific parameter.

    This class provides a user-friendly functionality which checks whether
    a certain parameter value falls within its preferred numerical range
    depending on its algorithm.

    Parameters
    ----------
    `param_range` : RangeStr
        An interval of a parameter (customizable)
    `param_type` : Type[Scalar], optional, default=None
        Data type of a parameter to be forced to have
        (None for both `int` and `float` types)

    Method
    ------
    To check its validity:
    ```py
    def check(self, param_value: Any) -> None
    ```
    Raises
    ------
    - `InvalidRangeError` :
        When the parameter does not fall in its preferred range
    - `UnsupportedParameterError` :
        When the parameter is not numeric or has wrong numeric type

    Examples
    --------
    ```py
    class SomeModel:
        def __init__(self, a: int) -> None:
            self.a = a
            a_range = ParamRange(param_range="0,+inf", param_type=int)
            a_range.check(param_value=self.a)
    ```
    Notes
    -----
    - When setting a custom range for `param_range`, it must follow the form of
        "lower_bound,upper_bound". (i.e. `-inf,20`, `-5,5`)
    - For open intervals, add '<' inside the range. (i.e. `0<,+inf`, `0<,<10`)
    """

    type RangeStr = str

    def __init__(
        self,
        param_range: RangeStr,
        param_type: Type[Scalar] = None,
    ) -> None:
        self.param_range = param_range

        if param_type is None:
            self.param_type = int | float
        else:
            if not ParamRange.validate_type(param_type):
                raise UnsupportedParameterError(param_type)
            self.param_type = param_type

    @classmethod
    def validate_type(cls, param_type) -> TypeGuard[Scalar]:
        return param_type in (int, float)

    def check(self, param_name: str, param_value: Any) -> None:
        if param_value is None:
            return
        if isinstance(param_value, (tuple, list)):
            for element in param_value:
                self._type_check(element)
        else:
            self._type_check(param_value)

        if not self.condition(param_value):
            raise InvalidRangeError(param_value, param_name, self.param_range)

    def _type_check(self, param_value: Any) -> None:
        if not isinstance(param_value, self.param_type):
            raise UnsupportedParameterError(param_value)

    @property
    def condition(self) -> Callable[[Scalar], bool]:
        try:
            left, right = self.param_range.split(",")
        except:
            raise UnsupportedParameterError(self.param_range)

        lower_open, upper_open = False, False
        if left[-1] == "<":
            lower_open = True
            lower = np.float64(left[:-1])
        else:
            lower = np.float64(left)

        if right[0] == "<":
            upper_open = True
            upper = np.float64(right[1:])
        else:
            upper = np.float64(right)

        if lower_open and not upper_open:
            return lambda x: lower < x <= upper
        elif lower_open and upper_open:
            return lambda x: lower < x < upper
        elif not lower_open and upper_open:
            return lambda x: lower <= x < upper
        elif not lower_open and not upper_open:
            return lambda x: lower <= x <= upper
        else:
            NotImplemented


class InitUtil:
    """
    An utility class for weight initializers used in neural networks.

    Parameters
    ----------
    `initializer` : InitStr, optional, default=True
        Name of an initializer

    Properties
    ----------
    To get the corresponding initializer type:
    ```py
    @property
    def initializer_type(self) -> type | None
    # `None` for random init
    ```
    """

    InitStr = Literal["he", "kaiming", "xavier", "glorot"] | None

    def __init__(self, initializer: InitStr) -> None:
        self.initializer = initializer

    @property
    def initializer_type(self) -> type | None:
        if self.initializer is None:
            return None
        if self.initializer in ("he", "kaiming"):
            return init.KaimingInit
        elif self.initializer in ("xavier", "glorot"):
            return init.XavierInit


class TrainProgress:
    """
    An utility class for managing auto-updating progress bar during training.

    It facilitates the progress bar from the module `rich`, which provides
    well-structured progress bar with colored indications for better readability.

    Parameters
    ----------
    `n_iter` : int
        Number of iterations(or epochs)
    `bar_width` : int, default=50
        Width size of a progress bar

    Property
    --------
    To get an iterable object from `rich.progress.Progress`:
    ```py
    (property) progress: (self: Self@TrainProgress) -> Iterable
    ```

    Examples
    --------
    Create an instance for `TrainProgress`:
    >>> train_prog = TrainProgress(n_iter=100)

    Generate a task and update the progress bar:
    ```py
    with train_prog.progress as progress:
        train_prog.add_task(progress=progress, model=AnyModelInstance)

        for i in range(n_iter):
            train_prog.update(progress=progress, cur=i, losses=[...])
            # losses is a list of [train_loss, valid_loss]
    ```
    """

    def __init__(self, n_iter: int, bar_width: int = 50) -> None:
        self.n_iter = n_iter
        self.bar_width = bar_width
        self.task = None

    @property
    def progress(self) -> Iterable:
        return Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(bar_width=self.bar_width),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            TextColumn("[bold green]{task.fields[train_loss]:.4f}"),
            TextColumn("[bold red]{task.fields[valid_loss]:.4f}"),
            expand=True,
        )

    def _check_task_exist(self) -> None:
        if self.task is None:
            raise RuntimeError(f"'{type(self).__name__}' does not have any tasks!")

    def add_task(self, progress: Progress, model: object) -> None:
        self.task = progress.add_task(
            f"[purple]Start {type(model).__name__} training "
            + f"with {self.n_iter} epochs.",
            total=self.n_iter,
            train_loss=0.0,
            valid_loss=0.0,
        )

    def update(
        self,
        progress: Progress,
        cur: int,
        losses: list[float, float],
    ) -> None:
        self._check_task_exist()
        progress.update(
            self.task,
            advance=1,
            train_loss=losses[0],
            valid_loss=losses[1],
            description=f"Epoch: {cur}/{self.n_iter} - "
            + f"Train/Valid Loss: {losses[0]:.4f}/{losses[1]:.4f}",
        )



from typing import Any, List, Literal, Self, Tuple, override

from luma.core.super import Optimizer
from luma.interface.typing import TensorLike
from luma.interface.util import InitUtil, Clone
from luma.neural.base import Layer

from luma.neural._layers import (
    _act,
    _conv,
    _drop,
    _linear,
    _norm,
    _pool,
)


__all__ = (
    "Convolution1D",
    "Convolution2D",
    "Convolution3D",
    "Pooling1D",
    "Pooling2D",
    "Pooling3D",
    "GlobalAvgPooling1D",
    "GlobalAvgPooling2D",
    "GlobalAvgPooling3D",
    "LpPooling1D",
    "LpPooling2D",
    "LpPooling3D",
    "Dense",
    "Dropout",
    "Dropout1D",
    "Dropout2D",
    "Dropout3D",
    "Flatten",
    "Activation",
    "BatchNorm1D",
    "BatchNorm2D",
    "BatchNorm3D",
    "LocalResponseNorm",
    "LayerNorm",
    "Sequential",
)


class Convolution1D(_conv._Conv1D):
    """
    Convolutional layer for 1-dimensional data.

    A convolutional layer in a neural network convolves learnable filters
    across input data, detecting patterns like edges or textures, producing
    feature maps essential for tasks such as image recognition within CNNs.
    By sharing parameters, it efficiently extracts hierarchical representations,
    enabling the network to learn complex visual features.

    Parameters
    ----------
    `in_channels` : int
        Number of input channels
    `out_channels` : int
        Number of output channels(filters)
    `filter_size`: tuple of int or int
        Size of each filter
    `stride` : int, default=1
        Step size for filters during convolution
    `padding` : tuple of int or int or {"valid", "same"}, default="same"
        Padding strategies ("valid" for no padding, "same" for zero-padding)
    `initializer` : InitStr, default=None
        Type of weight initializer
    `optimizer` : Optimizer, optional, default=None
        Optimizer for weight update
    `lambda_` : float, default=0.0
        L2-regularization strength
    `random_state` : int, optional, default=None
        Seed for various random sampling processes

    Notes
    -----
    - The input `X` must have the form of 3D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, width)
        ```
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        filter_size: Tuple[int] | int,
        stride: int = 1,
        padding: Tuple[int] | int | Literal["valid", "same"] = "same",
        initializer: InitUtil.InitStr = None,
        optimizer: Optimizer | None = None,
        lambda_: float = 0,
        random_state: int | None = None,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            filter_size,
            stride,
            padding,
            initializer,
            optimizer,
            lambda_,
            random_state,
        )


class Convolution2D(_conv._Conv2D):
    """
    Convolutional layer for 2-dimensional data.

    A convolutional layer in a neural network convolves learnable filters
    across input data, detecting patterns like edges or textures, producing
    feature maps essential for tasks such as image recognition within CNNs.
    By sharing parameters, it efficiently extracts hierarchical representations,
    enabling the network to learn complex visual features.

    Parameters
    ----------
    `in_channels` : int
        Number of input channels
    `out_channels` : int
        Number of output channels(filters)
    `filter_size`: tuple of int or int
        Size of each filter
    `stride` : int, default=1
        Step size for filters during convolution
    `padding` : tuple of int or int or {"valid", "same"}, default="same"
        Padding strategies ("valid" for no padding, "same" for zero-padding)
    `initializer` : InitStr, default=None
        Type of weight initializer
    `optimizer` : Optimizer, optional, default=None
        Optimizer for weight update
    `lambda_` : float, default=0.0
        L2-regularization strength
    `random_state` : int, optional, default=None
        Seed for various random sampling processes

    Notes
    -----
    - The input `X` must have the form of 4D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, height, width)
        ```
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        filter_size: Tuple[int, int] | int,
        stride: int = 1,
        padding: Tuple[int, int] | int | Literal["valid", "same"] = "same",
        initializer: InitUtil.InitStr = None,
        optimizer: Optimizer = None,
        lambda_: float = 0,
        random_state: int = None,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            filter_size,
            stride,
            padding,
            initializer,
            optimizer,
            lambda_,
            random_state,
        )


class Convolution3D(_conv._Conv3D):
    """
    Convolutional layer for 3-dimensional data.

    A convolutional layer in a neural network convolves learnable filters
    across input data, detecting patterns like edges or textures, producing
    feature maps essential for tasks such as image recognition within CNNs.
    By sharing parameters, it efficiently extracts hierarchical representations,
    enabling the network to learn complex visual features.

    Parameters
    ----------
    `in_channels` : int
        Number of input channels
    `out_channels` : int
        Number of output channels(filters)
    `filter_size`: tuple of int or int
        Size of each filter
    `stride` : int, default=1
        Step size for filters during convolution
    `padding` : tuple of int or int or {"valid", "same"}, default="same"
        Padding strategies ("valid" for no padding, "same" for zero-padding)
    `initializer` : InitStr, default=None
        Type of weight initializer
    `optimizer` : Optimizer, optional, default=None
        Optimizer for weight update
    `lambda_` : float, default=0.0
        L2-regularization strength
    `random_state` : int, optional, default=None
        Seed for various random sampling processes

    Notes
    -----
    - The input `X` must have the form of 5D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, depth, height, width)
        ```
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        filter_size: Tuple[int, int, int] | int,
        stride: int = 1,
        padding: Tuple[int, int, int] | int | Literal["valid", "same"] = "same",
        initializer: InitUtil.InitStr = None,
        optimizer: Optimizer = None,
        lambda_: float = 0,
        random_state: int = None,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            filter_size,
            stride,
            padding,
            initializer,
            optimizer,
            lambda_,
            random_state,
        )


class Pooling1D(_pool._Pool1D):
    """
    Pooling layer for 1-dimensional data.

    A pooling layer in a neural network reduces the spatial dimensions of
    feature maps, reducing computational complexity. It aggregates neighboring
    values, typically through operations like max pooling or average pooling,
    to extract dominant features. Pooling helps in achieving translation invariance
    and reducing overfitting by summarizing the presence of features in local
    regions. It downsamples feature maps, preserving important information while
    discarding redundant details.

    Parameters
    ----------
    `filter_size` : int, default=2
        Size of the pooling filter
    `stride` : int, default=2
        Step size of the filter during pooling
    `mode` : {"max", "avg"}, default="max"
        Pooling strategy (i.e., 'max' or 'avg')
    `padding` : tuple of int or int or {"valid", "same"}, default="valid"
        Padding strategies ("valid" for no padding, "same" for zero-padding)

    Notes
    -----
    - The input `X` must have the form of 3D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, width)
        ```
    """

    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        mode: Literal["max", "avg"] = "max",
        padding: Tuple[int] | int | Literal["valid", "same"] = "valid",
    ) -> None:
        super().__init__(filter_size, stride, mode, padding)


class Pooling2D(_pool._Pool2D):
    """
    Pooling layer for 2-dimensional data.

    A pooling layer in a neural network reduces the spatial dimensions of
    feature maps, reducing computational complexity. It aggregates neighboring
    values, typically through operations like max pooling or average pooling,
    to extract dominant features. Pooling helps in achieving translation invariance
    and reducing overfitting by summarizing the presence of features in local
    regions. It downsamples feature maps, preserving important information while
    discarding redundant details.

    Parameters
    ----------
    `filter_size` : int, default=2
        Size of the pooling filter
    `stride` : int, default=2
        Step size of the filter during pooling
    `mode` : {"max", "avg"}, default="max"
        Pooling strategy (i.e., 'max' or 'avg')
    `padding` : tuple of int or int or {"valid", "same"}, default="valid"
        Padding strategies ("valid" for no padding, "same" for zero-padding)

    Notes
    -----
    - The input `X` must have the form of 4D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, height, width)
        ```
    """

    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        mode: Literal["max", "avg"] = "max",
        padding: Tuple[int, int] | int | Literal["valid", "same"] = "valid",
    ) -> None:
        super().__init__(filter_size, stride, mode, padding)


class Pooling3D(_pool._Pool3D):
    """
    Pooling layer for 3-dimensional data.

    A pooling layer in a neural network reduces the spatial dimensions of
    feature maps, reducing computational complexity. It aggregates neighboring
    values, typically through operations like max pooling or average pooling,
    to extract dominant features. Pooling helps in achieving translation invariance
    and reducing overfitting by summarizing the presence of features in local
    regions. It downsamples feature maps, preserving important information while
    discarding redundant details.

    Parameters
    ----------
    `filter_size` : int, default=2
        Size of the pooling filter
    `stride` : int, default=2
        Step size of the filter during pooling
    `mode` : {"max", "avg"}, default="max"
        Pooling strategy (i.e., 'max' or 'avg')
    `padding` : tuple of int or int or {"valid", "same"}, default="valid"
        Padding strategies ("valid" for no padding, "same" for zero-padding)

    Notes
    -----
    - The input `X` must have the form of 5D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, depth, height, width)
        ```
    """

    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        mode: Literal["max", "avg"] = "max",
        padding: Tuple[int, int, int] | int | Literal["valid", "same"] = "valid",
    ) -> None:
        super().__init__(filter_size, stride, mode, padding)


class GlobalAvgPooling1D(_pool._GlobalAvgPool1D):
    """
    Global average pooling layer for 1-dimensional data.

    Global Average Pooling (GAP) is a downsampling technique in Convolutional
    Neural Networks (CNNs) that reduces each feature map to a single value by
    taking the average of all its elements. It effectively transforms a
    multi-dimensional feature map into a one-dimensional vector, which helps in
    reducing the number of parameters and avoiding overfitting.
    GAP is commonly used in classification tasks, particularly in the final layer
    before the output layer.

    Notes
    -----
    - The input `X` must have the form of 3D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, width)
        ```
    """

    def __init__(self) -> None:
        super().__init__()


class GlobalAvgPooling2D(_pool._GlobalAvgPool2D):
    """
    Global average pooling layer for 2-dimensional data.

    Global Average Pooling (GAP) is a downsampling technique in Convolutional
    Neural Networks (CNNs) that reduces each feature map to a single value by
    taking the average of all its elements. It effectively transforms a
    multi-dimensional feature map into a one-dimensional vector, which helps in
    reducing the number of parameters and avoiding overfitting.
    GAP is commonly used in classification tasks, particularly in the final layer
    before the output layer.

    Notes
    -----
    - The input `X` must have the form of 4D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, height, width)
        ```
    """

    def __init__(self) -> None:
        super().__init__()


class GlobalAvgPooling3D(_pool._GlobalAvgPool3D):
    """
    Global average pooling layer for 3-dimensional data.

    Global Average Pooling (GAP) is a downsampling technique in Convolutional
    Neural Networks (CNNs) that reduces each feature map to a single value by
    taking the average of all its elements. It effectively transforms a
    multi-dimensional feature map into a one-dimensional vector, which helps in
    reducing the number of parameters and avoiding overfitting.
    GAP is commonly used in classification tasks, particularly in the final layer
    before the output layer.

    Notes
    -----
    - The input `X` must have the form of 5D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, depth, height, width)
        ```
    """

    def __init__(self) -> None:
        super().__init__()


class LpPooling1D(_pool._LpPool1D):
    """
    Lp pooling layer for 1-dimensional data.

    Lp Pooling is a generalized pooling method where the pooling operation is
    based on the Lp norm. It computes the p-th power of the absolute values within
    a pooling window, sums them up, and then takes the p-th root of this sum.
    This approach provides a smooth transition between average pooling (p=1) and
    max pooling (p approaching infinity).

    Parameters
    ----------
    `filter_size` : int, default=2
        Size of the pooling filter
    `stride` : int, default=2
        Step size of the filter during pooling
    `p` : float, default=2.0
        Powering factor for Lp norm.
    `padding` : tuple of int or int or {"valid", "same"}, default="valid"
        Padding strategies ("valid" for no padding, "same" for zero-padding)

    Notes
    -----
    - The input `X` must have the form of 3D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, width)
        ```
    """

    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        p: float = 2,
        padding: Tuple[int] | int | Literal["same", "valid"] = "valid",
    ) -> None:
        super().__init__(filter_size, stride, p, padding)


class LpPooling2D(_pool._LpPool2D):
    """
    Lp pooling layer for 2-dimensional data.

    Lp Pooling is a generalized pooling method where the pooling operation is
    based on the Lp norm. It computes the p-th power of the absolute values within
    a pooling window, sums them up, and then takes the p-th root of this sum.
    This approach provides a smooth transition between average pooling (p=1) and
    max pooling (p approaching infinity).

    Parameters
    ----------
    `filter_size` : int, default=2
        Size of the pooling filter
    `stride` : int, default=2
        Step size of the filter during pooling
    `p` : float, default=2.0
        Powering factor for Lp norm.
    `padding` : tuple of int or int or {"valid", "same"}, default="valid"
        Padding strategies ("valid" for no padding, "same" for zero-padding)

    Notes
    -----
    - The input `X` must have the form of 4D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, height, width)
        ```
    """

    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        p: float = 2,
        padding: Tuple[int, int] | int | Literal["same", "valid"] = "valid",
    ) -> None:
        super().__init__(filter_size, stride, p, padding)


class LpPooling3D(_pool._LpPool3D):
    """
    Lp pooling layer for 3-dimensional data.

    Lp Pooling is a generalized pooling method where the pooling operation is
    based on the Lp norm. It computes the p-th power of the absolute values within
    a pooling window, sums them up, and then takes the p-th root of this sum.
    This approach provides a smooth transition between average pooling (p=1) and
    max pooling (p approaching infinity).

    Parameters
    ----------
    `filter_size` : int, default=2
        Size of the pooling filter
    `stride` : int, default=2
        Step size of the filter during pooling
    `p` : float, default=2.0
        Powering factor for Lp norm.
    `padding` : tuple of int or int or {"valid", "same"}, default="valid"
        Padding strategies ("valid" for no padding, "same" for zero-padding)

    Notes
    -----
    - The input `X` must have the form of 5D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, depth, height, width)
        ```
    """

    def __init__(
        self,
        filter_size: int = 2,
        stride: int = 2,
        p: float = 2,
        padding: Tuple[int, int, int] | int | Literal["same", "valid"] = "valid",
    ) -> None:
        super().__init__(filter_size, stride, p, padding)


class Dense(_linear._Dense):
    """
    A dense layer, also known as a fully connected layer, connects each
    neuron in one layer to every neuron in the next layer. It performs a
    linear transformation followed by a nonlinear activation function,
    enabling complex relationships between input and output. Dense layers
    are fundamental in deep learning models for learning representations from
    data. They play a crucial role in capturing intricate patterns and
    features during the training process.

    Parameters
    ----------
    `in_features` : int
        Number of input features
    `out_features`:  int
        Number of output features
    `initializer` : InitStr, default = None
        Type of weight initializer
    `optimizer` : Optimizer, optional, default=None
        Optimizer for weight update
    `lambda_` : float, default=0.0
        L2-regularization strength
    `random_state` : int, optional, default=None
        Seed for various random sampling processes

    Notes
    -----
    - The input `X` must have the form of 2D-array(`Matrix`).

        ```py
        X.shape = (batch_size, n_features)
        ```
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        initializer: InitUtil.InitStr = None,
        optimizer: Optimizer | None = None,
        lambda_: float = 0,
        random_state: int | None = None,
    ) -> None:
        super().__init__(
            in_features,
            out_features,
            initializer,
            optimizer,
            lambda_,
            random_state,
        )


class Dropout(_drop._Dropout):
    """
    Dropout is a regularization technique used during training to prevent
    overfitting by randomly setting a fraction of input units to zero during
    the forward pass. This helps in reducing co-adaptation of neurons and
    encourages the network to learn more robust features.

    Parameters
    ----------
    `dropout_rate` : float, default=0.5
        The fraction of input units to drop during training
    `random_state` : int, optional, default=None
        Seed for various random sampling processes

    Notes
    -----
    - This class applies dropout for every element in the input with any shape.
    """

    def __init__(
        self, dropout_rate: float = 0.5, random_state: int | None = None
    ) -> None:
        super().__init__(dropout_rate, random_state)


class Dropout1D(_drop._Dropout1D):
    """
    Dropout layer for 1-dimensional data.

    Dropout is a regularization technique used during training to prevent
    overfitting by randomly setting a fraction of input units to zero during
    the forward pass. This helps in reducing co-adaptation of neurons and
    encourages the network to learn more robust features.

    Parameters
    ----------
    `dropout_rate` : float, default=0.5
        The fraction of input units to drop during training
    `random_state` : int, optional, default=None
        Seed for various random sampling processes

    Notes
    -----
    - The input `X` must have the form of 3D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, width)
        ```
    """

    def __init__(
        self, dropout_rate: float = 0.5, random_state: int | None = None
    ) -> None:
        super().__init__(dropout_rate, random_state)


class Dropout2D(_drop._Dropout2D):
    """
    Dropout layer for 2-dimensional data.

    Dropout is a regularization technique used during training to prevent
    overfitting by randomly setting a fraction of input units to zero during
    the forward pass. This helps in reducing co-adaptation of neurons and
    encourages the network to learn more robust features.

    Parameters
    ----------
    `dropout_rate` : float, default=0.5
        The fraction of input units to drop during training
    `random_state` : int, optional, default=None
        Seed for various random sampling processes

    Notes
    -----
    - The input `X` must have the form of 4D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, height, width)
        ```
    """

    def __init__(
        self, dropout_rate: float = 0.5, random_state: int | None = None
    ) -> None:
        super().__init__(dropout_rate, random_state)


class Dropout3D(_drop._Dropout3D):
    """
    Dropout layer for 3-dimensional data.

    Dropout is a regularization technique used during training to prevent
    overfitting by randomly setting a fraction of input units to zero during
    the forward pass. This helps in reducing co-adaptation of neurons and
    encourages the network to learn more robust features.

    Parameters
    ----------
    `dropout_rate` : float, default=0.5
        The fraction of input units to drop during training
    `random_state` : int, optional, default=None
        Seed for various random sampling processes

    Notes
    -----
    - The input `X` must have the form of 5D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, depth, height, width)
        ```
    """

    def __init__(
        self, dropout_rate: float = 0.5, random_state: int | None = None
    ) -> None:
        super().__init__(dropout_rate, random_state)


class Flatten(_linear._Flatten):
    """
    A flatten layer reshapes the input tensor into a 2D array(`Matrix`),
    collapsing all dimensions except the batch dimension.

    Notes
    -----
    - Use this class when using `Dense` layer.
        Flatten the tensor into matrix in order to feed-forward dense layer(s).
    """

    def __init__(self) -> None:
        super().__init__()


class Activation(_act._Activation):
    """
    An Activation Layer in a neural network applies a specific activation
    function to the input it receives, transforming the input to activate
    or deactivate neurons within the network. This function can be linear
    or non-linear, such as Sigmoid, ReLU, or Tanh, which helps to introduce
    non-linearity into the model, allowing it to learn complex patterns.

    Notes
    -----
    - This class is not instantiable, meaning that a solitary use of it
        is not available.

    - All the activation functions are included inside `Activation`.

    Examples
    --------
    >>> # Activation() <- Impossible
    >>> Activation.Linear()
    >>> Activation.ReLU()

    """


class BatchNorm1D(_norm._BatchNorm1D):
    """
    Batch normalization layer for 1-dimensional data.

    Batch normalization standardizes layer inputs across mini-batches to stabilize
    learning, accelerate convergence, and reduce sensitivity to initialization.
    It adjusts normalized outputs using learnable parameters, mitigating internal
    covariate shift in deep networks.

    Parameters
    ----------
    `in_features` : int
        Number of input features
    `momentum` : float, default=0.9
        Momentum for updating the running averages

    Notes
    -----
    - The input `X` must have the form of 3D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, width)
        ```
    """

    def __init__(
        self, in_features: int, momentum: float = 0.9, epsilon: float = 1e-9
    ) -> None:
        super().__init__(in_features, momentum, epsilon)


class BatchNorm2D(_norm._BatchNorm2D):
    """
    Batch normalization layer for 2-dimensional data.

    Batch normalization standardizes layer inputs across mini-batches to stabilize
    learning, accelerate convergence, and reduce sensitivity to initialization.
    It adjusts normalized outputs using learnable parameters, mitigating internal
    covariate shift in deep networks.

    Parameters
    ----------
    `in_features` : int
        Number of input features
    `momentum` : float, default=0.9
        Momentum for updating the running averages

    Notes
    -----
    - The input `X` must have the form of 4D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, height, width)
        ```
    """

    def __init__(
        self, in_features: int, momentum: float = 0.9, epsilon: float = 1e-9
    ) -> None:
        super().__init__(in_features, momentum, epsilon)


class BatchNorm3D(_norm._BatchNorm3D):
    """
    Batch normalization layer for 3-dimensional data.

    Batch normalization standardizes layer inputs across mini-batches to stabilize
    learning, accelerate convergence, and reduce sensitivity to initialization.
    It adjusts normalized outputs using learnable parameters, mitigating internal
    covariate shift in deep networks.

    Parameters
    ----------
    `in_features` : int
        Number of input features
    `momentum` : float, default=0.9
        Momentum for updating the running averages

    Notes
    -----
    - The input `X` must have the form of 5D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, depth, height, width)
        ```
    """

    def __init__(
        self, in_features: int, momentum: float = 0.9, epsilon: float = 0.00001
    ) -> None:
        super().__init__(in_features, momentum, epsilon)


class LocalResponseNorm(_norm._LocalResponseNorm):
    """
    Local Response Normalization (LRN) is a technique used in neural networks
    to promote competition among neighboring feature maps. By normalizing the
    activities in local regions across channels, LRN helps to enhance generalization
    by suppressing activations that are uniformly large across the entire map and
    boosting those that are uniquely larger in a local neighborhood.

    Parameters
    ----------
    `depth` : int
        Number of adjacent channels to normalize across
    `alpha` : float, default=1e-4
        Scaling parameter for the squared sum
    `beta` : float, default=0.75
        Exponent for the normalization
    `k` : float, default=2
        Offset to avoid division by zero

    Notes
    -----
    - The input `X` must have the form of >=3D-array(`Tensor`).

        ```py
        X.shape = (batch_size, channels, *spatial)
        ```
    """

    def __init__(
        self,
        depth: int,
        alpha: float = 0.0001,
        beta: float = 0.75,
        k: float = 2,
    ) -> None:
        super().__init__(
            depth,
            alpha,
            beta,
            k,
        )


class LayerNorm(_norm._LayerNorm):
    """
    Layer normalization is a technique used in neural networks to normalize the
    inputs across the features for each data sample in a batch independently.
    It stabilizes the learning process by reducing the variance of the inputs
    within a layer, helping to speed up training and improve performance.

    Parameters
    ----------
    `in_shape` : int or tuple of int
        Shape of the input

    Notes
    -----
    - The input `X` must have the form of >=2D-array(`Matrix`).

        ```py
        X.shape = (batch_size, *spatial)
        ```
    """

    def __init__(
        self,
        in_shape: Tuple[int] | int,
        epsilon: float = 1e-5,
    ) -> None:
        super().__init__(in_shape, epsilon)


type LayerLike = Sequential | Layer


class Sequential(Layer):
    """
    Sequential represents a linear arrangement of layers in a neural network
    model. Each layer is added sequentially, with data flowing from one layer
    to the next in the order they are added. This organization simplifies the
    construction of neural networks, especially for straightforward architectures,
    by mirroring the logical flow of data from input through hidden layers to
    output.

    Parameters
    ----------
    `*layers` : Layer or tuple[str, Layer], optional
        Layers or layers with its name assigned
        (class name of the layer assigned by default)

    Methods
    -------
    For setting an optimizer of each layer:
    ```py
    def set_optimizer(self, optimizer: Optimizer) -> None
    ```
    To add additional layer:
    ```py
    def add(self, layer: Layer) -> None
    ```
    Specials
    --------
    - You can use `+` operator to add a layer or another instance
        of `Sequential`.

    - Calling its instance performs a single forwarding.

    Notes
    -----
    - Before any execution, an optimizer must be assigned.

    - For multi-class classification, the target variable `y`
        must be one-hot encoded.

    Examples
    --------
    ```py
    model = Sequential(
        ("conv_1", Convolution(3, 6, 3, activation="relu")),
        ("pool_1", Pooling(2, 2, mode="max")),
        ...,
        ("drop", Dropout(0.1)),
        ("flat", Flatten()),
        ("dense_1", Dense(384, 32, activation="relu")),
        ("dense_2", Dense(32, 10, activation="softmax")),
    )
    model.set_optimizer(AnyOptimizer())

    out = model(X, is_train=True) # model.forward(X, is_train=True)
    model.backward(d_out) # assume d_out is the gradient w.r.t. loss
    model.update()
    ```
    """

    def __init__(self, *layers: LayerLike | tuple[str, LayerLike] | None) -> None:
        super().__init__()
        self.layers: List[tuple[str, LayerLike]] = list()
        for layer in layers:
            self.add(layer)

    def forward(self, X: TensorLike, is_train: bool = False) -> TensorLike:
        self.input_ = X
        out = X
        for _, layer in self.layers:
            out = layer(out, is_train=is_train)

        return out

    def backward(self, d_out: TensorLike) -> TensorLike:
        for _, layer in reversed(self.layers):
            d_out = layer.backward(d_out)
        return d_out

    def update(self) -> None:
        self._check_no_optimizer()
        for _, layer in reversed(self.layers):
            layer.update()

    def set_optimizer(self, optimizer: Optimizer, **params: Any) -> None:
        self.optimizer = optimizer
        self.optimizer.set_params(**params, ignore_missing=True)

        for _, layer in self.layers:
            cloned_opt = Clone(self.optimizer).get
            if hasattr(layer, "set_optimizer"):
                layer.set_optimizer(cloned_opt)
            else:
                layer.optimizer = cloned_opt

    def _check_no_optimizer(self) -> None:
        if self.optimizer is None:
            raise RuntimeError(
                f"'{self}' has no optimizer! "
                + f"Call '{self}().set_optimizer' to assign an optimizer."
            )

    def add(self, layer: LayerLike | tuple[str, LayerLike] | None) -> None:
        if layer is None:
            return
        if not isinstance(layer, tuple):
            layer = (str(layer), layer)
        self.layers.append(layer)

        if self.optimizer is not None:
            self.set_optimizer(self.optimizer)

    def extend(
        self,
        *layers: LayerLike | tuple[str, LayerLike] | None,
        deep_add: bool = True,
    ) -> None:
        for layer in layers:
            new_layer = layer
            if isinstance(layer, tuple):
                name, layer = layer
                new_layer = (name, layer)
            if hasattr(layer, "layers") and deep_add:
                for sub_layer in layer.layers:
                    self.add(sub_layer)
                continue
            self.add(new_layer)

    def override_method(self, func_name: str, func: callable) -> None:
        if not hasattr(self, func_name):
            raise RuntimeError(f"'{str(self)}' has no method called '{func_name}'!")
        if not callable(getattr(self, func_name)):
            raise TypeError(f"'{func_name}' it not a method!")
        if not callable(func):
            raise TypeError(f"Provided method '{func}' it not a method!")

        setattr(self, func_name, func)

    @override
    @property
    def param_size(self) -> Tuple[int, int]:
        w_size, b_size = 0, 0
        for _, layer in self.layers:
            w_, b_ = layer.param_size
            w_size += w_
            b_size += b_

        return w_size, b_size

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        for _, layer in self.layers:
            in_shape = layer.out_shape(in_shape)
        return in_shape

    def __add__(self, other: LayerLike | tuple[str, LayerLike] | None) -> Self:
        if isinstance(other, (Layer, tuple)):
            self.add(other)
        else:
            raise TypeError(
                "Unsupported operand type(s) for +: '{}' and '{}'".format(
                    type(self).__name__, type(other).__name__
                )
            )

        return self

    def __getitem__(self, index: int) -> Tuple[str, Layer]:
        return self.layers[index]



from functools import wraps
from typing import Any, Callable, Generic, NoReturn, Self, Type, TypeVar
import sys
import numpy as np


__all__ = (
    "TensorLike",
    "Matrix",
    "Vector",
    "Tensor",
    "Scalar",
    "ClassType",
)


T = TypeVar("T", bound=type)
D = TypeVar("D", bound=int)


class TensorLike(np.ndarray):
    """
    Internal base class for n-dimensional arrays that extends
    `numpy.ndarray`.

    Apart from the type `Tensor`, this class encompasses all the
    mathematical notion of generalized tensors, including
    vectors(1D-array), matrices(2D-array).
    """

    def __init__(self) -> None:
        super().__init__()

    def __array_finalize__(self, obj: None | np.ndarray[Any, np.dtype[Any]]) -> None:
        return super().__array_finalize__(obj)

    def __init_subclass__(cls) -> None:
        return super().__init_subclass__()


class Matrix(TensorLike):
    """
    Internal class for matrices(2D-array) that extends `TensorLike`.

    This class provides a way to create matrix objects that have
    all the capabilities of numpy arrays with the potential for
    additional functionalities and readability.

    Example
    -------
    >>> m = Matrix([1, 2, 3])

    """

    def __new__(cls, array_like: Any) -> Self:
        if isinstance(array_like, (list, np.matrix)):
            obj = np.array(array_like)
        else:
            obj = array_like
        return obj

    def __array_finalize__(self, obj: np.ndarray | None) -> None:
        if obj is None:
            return


class Vector(TensorLike):
    """
    Internal class for vectors(1D-array) that extends `TensorLike`.

    This class represents a single row/column vector with its
    type of `TensorLike`.

    """

    def __new__(cls, array_like: Any) -> Self:
        if isinstance(array_like, list):
            obj = Matrix(array_like)
        else:
            obj = array_like
        return obj


class Tensor(TensorLike, Generic[D]):
    """
    Internal class for tensors(>=3D-arrray) that extends `TensorLike`.

    This class provides a way to create tensor objects that have
    all the capabilities of numpy arrays with the potential for
    additional functionalities and readability.

    Notes
    -----
    In general mathematics and physics, a term 'Tensor' refers to a
    generalized n-dimensional arrays, whereas in `luma`, `Tensor` only
    refers to an array with its dimensionality higher then 2.
    """

    def __new__(cls, array_like: Any) -> Self:
        if isinstance(array_like, list):
            obj = Matrix(array_like)
        else:
            obj = array_like
        return obj

    @classmethod
    def force_dim(cls, *dim_consts: int) -> Callable:
        """
        Decorator factory to enforce the dimensionality of tensor arguments.

        This decorator ensures that specified tensor arguments to the decorated
        function have the expected number of dimensions.

        Parameters
        ----------
        `*dim_constraints` : int
            Variable length argument list where each value specifies the required
            number of dimensions for the corresponding tensor argument in the
            decorated function.

        Returns
        -------
        `Callable`
            The decorated function with dimensionality constraints enforced on
            specified tensor arguments.

        Raises
        ------
        `TypeError`
            If the argument is not a numpy ndarray.
        `ValueError`
            If the argument does not have the specified number of dimensions.

        Examples
        --------
        >>> class MyModel:
        ...     @Tensor.force_dim(4)
        ...     def foo(self, tensor: Tensor) -> Any:
        ...         print("Processing foo")
        ...
        >>> model = MyModel()
        >>> tensor_4d = np.random.rand(2, 3, 4, 5).view(Tensor)
        >>> model.forward(tensor_4d)
        Processing foo

        >>> tensor_3d = np.random.rand(2, 3, 4).view(Tensor)
        >>> model.foo(tensor_3d)
        Traceback (most recent call last): ...
        ValueError: 'X' must be 4-dimensional, got 3 dimensions

        """

        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            @wraps(func)
            def wrapper(self, *args: Any, **kwargs: Any) -> Any:
                arg_names = func.__code__.co_varnames
                all_args = {**dict(zip(arg_names, (self,) + args)), **kwargs}

                for i, n_dim in enumerate(dim_consts):
                    param_name = arg_names[i + 1]

                    if param_name in all_args:
                        tensor = all_args[param_name]
                        if not isinstance(tensor, (Tensor, np.ndarray)):
                            raise TypeError(f"'{param_name}' must be of type Tensor.")
                        if tensor.ndim != n_dim:
                            raise ValueError(
                                f"'{param_name}' must be {n_dim}D-tensor",
                                +f" got {tensor.ndim}D-tensor.",
                            )

                return func(self, *args, **kwargs)

            return wrapper

        return decorator


class Scalar:
    """
    A placeholder class for scalar type.

    This class encompasses `int` and `float`.
    """

    def __new__(cls, value: int | float) -> Self:
        return value


class ClassType:
    """
    An interface class designed to facilitate the annotation of
    dedicated types for classes in a systematic way.

    The `ClassType` class includes several classmethods that act
    as decorator factories. These decorators can be applied to classes
    to specify their types or roles within the application more
    explicitly.

    Decorator Factories
    -------------------
    `non_instantiable` : callable
        Makes a class non-instantiable.
    `private` : callable
        Makes a class private; accesses are restricted outside its module.

    """

    @classmethod
    def non_instantiable(cls) -> Callable:
        def decorator(cls: Type[T]) -> Type[T]:
            def wrapper(*args, **kwargs) -> NoReturn:
                args, kwargs
                raise TypeError(
                    f"'{cls.__name__}'" + " is not instantiable!",
                )

            cls.__new__ = wrapper
            return cls

        return decorator

    @classmethod
    def private(cls) -> Callable:
        def decorator(cls: Type[T]) -> Type[T]:

            @wraps(cls, updated=())
            class PrivateClassWrapper(cls):
                def __new__(cls, *args, **kwargs) -> Any:
                    args, kwargs
                    caller = sys._getframe(1)

                    if caller.f_globals["__name__"] == cls.__module__:
                        return super().__new__(cls)
                    else:
                        raise TypeError(
                            f"'{cls.__name__}' is a private class and cannot "
                            + "be instantiated outside its module.",
                        )

                def __init__(self, *args, **kwargs) -> None:
                    super(PrivateClassWrapper, self).__init__(*args, **kwargs)

            for attr in dir(cls):
                if not attr.startswith("_"):
                    setattr(PrivateClassWrapper, attr, getattr(cls, attr))

            return PrivateClassWrapper

        return decorator



from typing import Tuple
import numpy as np

from luma.interface.typing import Tensor
from luma.neural.base import Layer


__all__ = ("_Dropout", "_Dropout1D", "_Dropout2D", "_Dropout3D")


class _Dropout(Layer):
    def __init__(
        self,
        dropout_rate: float = 0.5,
        random_state: int | None = None,
    ) -> None:
        super(_Dropout, self).__init__()
        self.dropout_rate = dropout_rate
        self.random_state = random_state
        self.rs_ = np.random.RandomState(self.random_state)
        self.mask_ = None

        self.set_param_ranges({"dropout_rate": ("0,1", None)})
        self.check_param_ranges()

    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        if is_train:
            self.mask_ = self.rs_.rand(*X.shape) < (1 - self.dropout_rate)
            return X * self.mask_ / (1 - self.dropout_rate)
        return X

    def backward(self, d_out: Tensor) -> Tensor:
        if self.mask_ is not None:
            return d_out * self.mask_ / (1 - self.dropout_rate)
        return d_out

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        return in_shape


class _Dropout1D(Layer):
    def __init__(
        self,
        dropout_rate: float = 0.5,
        random_state: int | None = None,
    ) -> None:
        super().__init__()
        self.dropout_rate = dropout_rate
        self.random_state = random_state
        self.rs_ = np.random.RandomState(self.random_state)
        self.mask_ = None

        self.set_param_ranges({"dropout_rate": ("0,1", None)})
        self.check_param_ranges()

    @Tensor.force_dim(3)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        if is_train:
            self.mask_ = self.rs_.rand(*X.shape[:2], 1) < (1 - self.dropout_rate)
            return X * self.mask_ / (1 - self.dropout_rate)
        return X

    @Tensor.force_dim(3)
    def backward(self, d_out: Tensor) -> Tensor:
        if self.mask_ is not None:
            return d_out * self.mask_ / (1 - self.dropout_rate)
        return d_out

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        return in_shape


class _Dropout2D(Layer):
    def __init__(
        self,
        dropout_rate: float = 0.5,
        random_state: int | None = None,
    ) -> None:
        super().__init__()
        self.dropout_rate = dropout_rate
        self.random_state = random_state
        self.rs_ = np.random.RandomState(self.random_state)
        self.mask_ = None

        self.set_param_ranges({"dropout_rate": ("0,1", None)})
        self.check_param_ranges()

    @Tensor.force_dim(4)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        if is_train:
            self.mask_ = self.rs_.rand(*X.shape[:2], 1, 1) < (1 - self.dropout_rate)
            return X * self.mask_ / (1 - self.dropout_rate)
        return X

    @Tensor.force_dim(4)
    def backward(self, d_out: Tensor) -> Tensor:
        if self.mask_ is not None:
            return d_out * self.mask_ / (1 - self.dropout_rate)
        return d_out

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        return in_shape


class _Dropout3D(Layer):
    def __init__(
        self,
        dropout_rate: float = 0.5,
        random_state: int | None = None,
    ) -> None:
        super().__init__()
        self.dropout_rate = dropout_rate
        self.random_state = random_state
        self.rs_ = np.random.RandomState(self.random_state)
        self.mask_ = None

        self.set_param_ranges({"dropout_rate": ("0,1", None)})
        self.check_param_ranges()

    @Tensor.force_dim(5)
    def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
        if is_train:
            self.mask_ = self.rs_.rand(*X.shape[:2], 1, 1, 1) < (1 - self.dropout_rate)
            return X * self.mask_ / (1 - self.dropout_rate)
        return X

    @Tensor.force_dim(5)
    def backward(self, d_out: Tensor) -> Tensor:
        if self.mask_ is not None:
            return d_out * self.mask_ / (1 - self.dropout_rate)
        return d_out

    def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
        return in_shape



from typing import Literal

from luma.core.super import Optimizer
from luma.interface.util import InitUtil

from luma.neural import loss
from luma.neural.base import Loss
from luma.neural.layer import Activation

from luma.neural._models import _simple, _lenet, _alex, _vgg, _inception


__all__ = (
    "SimpleMLP",
    "SimpleCNN",
    "LeNet_1",
    "LeNet_4",
    "LeNet_5",
    "AlexNet",
    "ZFNet",
    "VGGNet_11",
    "VGGNet_13",
    "VGGNet_16",
    "VGGNet_19",
    "Inception_V1",
    "Inception_V2",
    "Inception_V3",
)


class SimpleMLP(_simple._SimpleMLP):
    """
    An MLP (Multilayer Perceptron) is a type of artificial neural network
    composed of at least three layers: an input layer, one or more hidden
    layers, and an output layer. Each layer consists of nodes, or neurons,
    which are fully connected to the neurons in the next layer. MLPs use a
    technique called backpropagation for learning, where the output error
    is propagated backwards through the network to update the weights.
    They are capable of modeling complex nonlinear relationships between
    inputs and outputs. MLPs are commonly used for tasks like classification,
    regression, and pattern recognition.

    Structure
    ---------
    ```py
    (Dense -> Activation -> Dropout) -> ... -> Dense
    ```
    Parameters
    ----------
    `in_features` : int
        Number of input features
    `out_features` : int
        Number of output features
    `hidden_layers` : int of list of int
        Numbers of the features in hidden layers (int for a single layer)
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.001
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `initializer` : InitStr, default=None
        Type of weight initializer
    `activation` : FuncType
        Type of activation function
    `optimizer` : Optimizer
        An optimizer used in weight update process
    `loss` : Loss
        Type of loss function
    `dropout_rate` : float, default=0.5
        Dropout rate
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    Notes
    -----
    - If the data or the target is a 1D-Array(`Vector`), reshape it into a
        higher dimensional array.

    - For classification tasks, the target vector `y` must be
        one-hot encoded.

    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        hidden_layers: list[int] | int,
        *,
        activation: Activation.FuncType,
        optimizer: Optimizer,
        loss: Loss,
        initializer: InitUtil.InitStr = None,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.001,
        valid_size: float = 0.1,
        dropout_rate: float = 0.5,
        lambda_: float = 0,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False
    ) -> None:
        super().__init__(
            in_features,
            out_features,
            hidden_layers,
            activation,
            optimizer,
            loss,
            initializer,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            dropout_rate,
            lambda_,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class SimpleCNN(_simple._SimpleCNN):
    """
    A Convolutional Neural Network (CNN) is a type of deep neural network
    primarily used in image recognition and processing that is particularly
    powerful at capturing spatial hierarchies in data. A CNN automatically
    detects important features without any human supervision using layers
    with convolving filters that pass over the input image and compute outputs.
    These networks typically include layers such as convolutional layers,
    pooling layers, and fully connected layers that help in reducing the
    dimensions while retaining important features.

    Structure
    ---------
    ```py
    ConvBlock2D -> ... -> Flatten -> DenseBlock -> ... -> Dense
    ```
    Parameters
    ----------
    `in_channels_list` : int or list of int
        List of input channels for convolutional blocks
    `in_features_list` : int or list of int
        List of input features for dense blocks
    `out_channels` : int
        Output channels for the last convolutional layer
    `out_features` : int
        Output features for the last dense layer
    `filter_size` : int
        Size of filters for convolution layers
    `activation` : FuncType
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer (None for dense layers)
    `padding` : {"same", "valid"}, default="same"
        Padding strategy
    `stride` : int, default=1
        Step size of filters during convolution
    `do_batch_norm` : bool, default=True
        Whether to perform batch normalization
    `momentum` : float, default=0.9
        Momentum for batch normalization
    `do_pooling` : bool, default=True
        Whether to perform pooling
    `pool_filter_size` : int, default=2
        Size of filters for pooling layers
    `pool_stride` : int, default=2
        Step size of filters during pooling
    `pool_mode` : {"max", "avg"}, default="max"
        Pooling strategy (default `max`)
    `do_dropout` : bool, default=True
        Whether to perform dropout
    `dropout_rate` : float, default=0.5
        Dropout rate
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.001
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    Notes
    -----
    - Input `X` must have the shape of 4D-array(`Tensor`)

    - For classification tasks, the target vector `y` must be
        one-hot encoded.

    Examples
    --------
    ```py
    model = SimpleCNN(
        in_channels_list=[1, 6],
        in_features_list=[96, 16],
        out_channels=12,
        out_features=10,
        activation=Activation.ReLU,
        ...,
    )
    ```
    This model has the same structure with:
    ```py
    model = Sequential(
        Convolution(1, 6),  # First convolution block
        Activation.ReLU,
        Pooling(),

        Convolution(6, 12),  # Second convolution block
        Activation.ReLU,
        Pooling(),

        Flatten(),
        Dense(96, 16),  # Dense block
        Activation.ReLU,
        Dropout(),

        Dense(16, 10),  # Final dense layer
    )
    ```
    """

    def __init__(
        self,
        in_channels_list: list[int] | int,
        in_features_list: list[int] | int,
        out_channels: int,
        out_features: int,
        *,
        filter_size: int,
        activation: Activation.FuncType,
        optimizer: Optimizer,
        loss: Loss,
        initializer: InitUtil.InitStr = None,
        padding: Literal["same", "valid"] = "same",
        stride: int = 1,
        do_batch_norm: bool = True,
        momentum: float = 0.9,
        do_pooling: bool = True,
        pool_filter_size: int = 2,
        pool_stride: int = 2,
        pool_mode: Literal["max", "avg"] = "max",
        do_dropout: bool = True,
        dropout_rate: float = 0.5,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.001,
        valid_size: float = 0.1,
        lambda_: float = 0,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False
    ) -> None:
        super().__init__(
            in_channels_list,
            in_features_list,
            out_channels,
            out_features,
            filter_size,
            activation,
            optimizer,
            loss,
            initializer,
            padding,
            stride,
            do_batch_norm,
            momentum,
            do_pooling,
            pool_filter_size,
            pool_stride,
            pool_mode,
            do_dropout,
            dropout_rate,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class LeNet_1(_lenet._LeNet_1):
    """
    LeNet-1 is an early convolutional neural network (CNN) proposed by
    Yann LeCun in 1988, primarily designed for handwritten character
    recognition. It consists of two convolutional layers interleaved
    with subsampling layers, followed by a fully connected layer.
    The network uses convolutions to automatically learn spatial
    hierarchies of features, which are then used for classification
    tasks. LeNet-1 was one of the first successful applications of CNNs,
    laying the groundwork for more complex architectures in image
    processing.

    Structure
    ---------
    Input:
    ```py
    Tensor[..., 1, 28, 28]
    ```
    Convolution Layers:
    ```py
    ConvBlock2D(1, 4) -> ConvBlock2D(4, 8)
    ```
    Fully Connected Layers:
    ```py
    Flatten -> Dense(8 * 4 * 4, 10)
    ```
    Output:
    ```py
    Matrix[..., 10]
    ```
    Parameter Size:
    ```txt
    2,180 weights, 22 biases -> 2,202 params
    ```
    Parameters
    ----------
    `activation` : FuncType, default=Activation.Tanh
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss, default=CrossEntropy()
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer
    `out_features` : int, default=10
        Number of output features
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.001
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    References
    ----------
    1. LeCun, Yann, et al. "Backpropagation Applied to Handwritten Zip
    Code Recognition." Neural Computation, vol. 1, no. 4, 1989, pp. 541-551.

    """

    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.Tanh,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 10,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False,
    ) -> None:
        super().__init__(
            optimizer,
            activation,
            loss,
            initializer,
            out_features,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class LeNet_4(_lenet._LeNet_4):
    """
    LeNet-4 is a specific convolutional neural network structure designed
    for more advanced image recognition tasks than its predecessors.
    This version incorporates several layers of convolutions and pooling,
    followed by fully connected layers leading to the output for classification.

    Structure
    ---------
    Input:
    ```py
    Tensor[..., 1, 32, 32]
    ```
    Convolution Layers:
    ```py
    ConvBlock2D(1, 4) -> ConvBlock2D(4, 16)
    ```
    Fully Connected Layers:
    ```py
    Flatten -> DenseBlock(16 * 5 * 5, 120) -> Dense(10)
    ```
    Output:
    ```py
    Matrix[..., 10]
    ```
    Parameter Size:
    ```txt
    50,902 weights, 150 biases -> 51,052 params
    ```
    Parameters
    ----------
    `activation` : FuncType, default=Activation.Tanh
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss, default=CrossEntropy()
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer
    `out_features` : int, default=10
        Number of output features
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.001
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    References
    ----------
    1. LeCun, Yann, et al. "Backpropagation Applied to Handwritten Zip
    Code Recognition." Neural Computation, vol. 1, no. 4, 1989, pp. 541-551.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.Tanh,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 10,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False,
    ) -> None:
        super().__init__(
            optimizer,
            activation,
            loss,
            initializer,
            out_features,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            dropout_rate,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class LeNet_5(_lenet._LeNet_5):
    """
    LeNet-5 is a specific convolutional neural network structure designed
    for more advanced image recognition tasks than its predecessors.
    This version incorporates several layers of convolutions and pooling,
    followed by fully connected layers leading to the output for classification.

    Structure
    ---------
    Input:
    ```py
    Tensor[..., 1, 32, 32]
    ```
    Convolution Layers:
    ```py
    ConvBlock2D(1, 6) -> ConvBlock2D(6, 16)
    ```
    Fully Connected Layers:
    ```py
    Flatten ->
    DenseBlock(16 * 5 * 5, 120) -> DenseBlock(120, 84) -> Dense(84, 10)
    ```
    Output:
    ```py
    Matrix[..., 10]
    ```
    Parameter Size:
    ```txt
    61,474 weights, 236 biases -> 61,710 params
    ```
    Parameters
    ----------
    `activation` : FuncType, default=Activation.Tanh
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss, default=CrossEntropy()
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer
    `out_features` : int, default=10
        Number of output features
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.001
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    References
    ----------
    1. LeCun, Yann, et al. "Backpropagation Applied to Handwritten Zip
    Code Recognition." Neural Computation, vol. 1, no. 4, 1989, pp. 541-551.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.Tanh,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 10,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False,
    ) -> None:
        super().__init__(
            optimizer,
            activation,
            loss,
            initializer,
            out_features,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            dropout_rate,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class AlexNet(_alex._AlexNet):
    """
    AlexNet is a deep convolutional neural network that is designed for
    challenging image recognition tasks and was the winning entry in ILSVRC 2012.
    This architecture uses deep layers of convolutions with ReLU activations,
    max pooling, dropout, and fully connected layers leading to a classification
    output.

    Structure
    ---------
    Input:
    ```py
    Tensor[..., 3, 227, 227]
    ```
    Convolutional Blocks:
    ```py
    ConvBlock2D(3, 96) -> LocalResponseNorm(5) ->  # Conv_1
    ConvBlock2D(96, 256) -> LocalResponseNorm(5) ->  # Conv_2

    ConvBlock2D(256, 384, do_pooling=False) -> LocalResponseNorm(5) ->  # Conv_3
    ConvBlock2D(384, 384, do_pooling=False) -> LocalResponseNorm(5) ->  # Conv_4

    ConvBlock2D(384, 256) -> LocalResponseNorm(5) ->  # Conv_5
    ```
    Fully Connected Layers:
    ```py
    Flatten ->
    DenseBlock(256 * 6 * 6, 4096) -> DenseBlock(4096, 4096) ->
    Dense(4096, 1000)
    ```
    Output:
    ```py
    Matrix[..., 1000]
    ```
    Parameter Size:
    ```txt
    62,367,776 weights, 10,568 biases -> 62,378,344 params
    ```
    Parameters
    ----------
    `activation` : FuncType, default=Activation.ReLU
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss, default=CrossEntropy()
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer
    `out_features` : int, default=1000
        Number of output features
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.01
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    References
    ----------
    1. Krizhevsky, Alex, Ilya Sutskever, and Geoffrey E. Hinton. "ImageNet
    Classification with Deep Convolutional Neural Networks." Advances in Neural
    Information Processing Systems, 2012.

    """

    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.ReLU,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 1000,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False,
    ) -> None:
        super().__init__(
            optimizer,
            activation,
            loss,
            initializer,
            out_features,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            dropout_rate,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class ZFNet(_alex._ZFNet):
    """
    ZFNet is a refinement of the AlexNet architecture that was specifically
    designed to improve model understanding and performance on image recognition
    tasks. This model was presented by Matthew Zeiler and Rob Fergus in their
    paper and was particularly notable for its improvements in layer configurations
    that enhanced visualization of intermediate activations, aiding in understanding
    the functioning of deep convolutional networks.

    Structure
    ---------
    Input:
    ```py
    Tensor[..., 3, 227, 227]
    ```
    Convolutional Blocks:
    ```py
    ConvBlock2D(3, 96) -> LocalResponseNorm(5) ->  # Conv_1
    ConvBlock2D(96, 256) -> LocalResponseNorm(5) ->  # Conv_2

    ConvBlock2D(256, 384, do_pooling=False) -> LocalResponseNorm(5) ->  # Conv_3
    ConvBlock2D(384, 384, do_pooling=False) -> LocalResponseNorm(5) ->  # Conv_4

    ConvBlock2D(384, 256) -> LocalResponseNorm(5) ->  # Conv_5
    ```
    Fully Connected Layers:
    ```py
    Flatten ->
    DenseBlock(256 * 6 * 6, 4096) -> DenseBlock(4096, 4096) ->
    Dense(4096, 1000)
    ```
    Output:
    ```py
    Matrix[..., 1000]
    ```
    Parameter Size:
    ```txt
    58,292,000 weights, 9,578 biases -> 58,301,578 params
    ```
    Parameters
    ----------
    `activation` : FuncType, default=Activation.ReLU
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss, default=CrossEntropy()
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer
    `out_features` : int, default=1000
        Number of output features
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.01
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    References
    ----------
    1. Zeiler, Matthew D., and Rob Fergus. "Visualizing and Understanding
    Convolutional Networks." European conference on computer vision, 2014.

    """

    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.ReLU,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 1000,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False,
    ) -> None:
        super().__init__(
            optimizer,
            activation,
            loss,
            initializer,
            out_features,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            dropout_rate,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class VGGNet_11(_vgg._VGGNet_11):
    """
    VGG11 is a simplified variant of the VGG network architecture that was designed
    to enhance image recognition performance through deeper networks with smaller
    convolutional filters. This model was introduced by Karen Simonyan and Andrew
    Zisserman in their paper and is notable for its simplicity and effectiveness
    in image classification tasks.

    Structure
    ---------
    Input:
    ```py
    Tensor[..., 3, 224, 224]
    ```
    Convolutional Blocks:
    ```py
    ConvBlock2D(3, 64) -> Pooling2D(2, 2, "max") ->  # Conv_1
    ConvBlock2D(64, 128) -> Pooling2D(2, 2, "max") ->  # Conv_2

    ConvBlock2D(128, 256, do_pooling=False) ->  # Conv_3
    ConvBlock2D(256, 256) -> Pooling2D(2, 2, "max") ->  # Conv_4

    ConvBlock2D(256, 512, do_pooling=False) ->  # Conv_5
    ConvBlock2D(512, 512) -> Pooling2D(2, 2, "max") ->  # Conv_6

    ConvBlock2D(512, 512, do_pooling=False) ->  # Conv_7
    ConvBlock2D(512, 512) -> Pooling2D(2, 2, "max") ->  # Conv_8
    ```
    Fully Connected Layers:
    ```py
    Flatten ->
    DenseBlock(512 * 7 * 7, 4096) -> DenseBlock(4096, 4096) ->
    Dense(4096, 1000)
    ```
    Output:
    ```py
    Matrix[..., 1000]
    ```
    Parameter Size:
    ```txt
    132,851,392 weights, 11,944 biases -> 132,863,336 params
    ```
    Parameters
    ----------
    `activation` : FuncType, default=Activation.ReLU
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss, default=CrossEntropy()
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer
    `out_features` : int, default=1000
        Number of output features
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.01
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    References
    ----------
    1. Simonyan, Karen, and Andrew Zisserman. "Very Deep Convolutional Networks for
    Large-Scale Image Recognition." arXiv preprint arXiv:1409.1556, 2014.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.ReLU,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 1000,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False,
    ) -> None:
        super().__init__(
            optimizer,
            activation,
            loss,
            initializer,
            out_features,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            dropout_rate,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class VGGNet_13(_vgg._VGGNet_13):
    """
    VGG13 is ont of the variants of the VGG network architecture that was designed
    to enhance image recognition performance through deeper networks with smaller
    convolutional filters. This model was introduced by Karen Simonyan and Andrew
    Zisserman in their paper and is notable for its simplicity and effectiveness
    in image classification tasks.

    Structure
    ---------
    Input:
    ```py
    Tensor[..., 3, 224, 224]
    ```
    Convolutional Blocks:
    ```py
    ConvBlock2D(3, 64, do_pooling=False) ->  # Conv_1
    ConvBlock2D(64, 64) -> Pooling2D(2, 2, "max") ->  # Conv_2

    ConvBlock2D(64, 128, do_pooling=False) ->  # Conv_3
    ConvBlock2D(128, 128) -> Pooling2D(2, 2, "max") ->  # Conv_4

    ConvBlock2D(128, 256, do_pooling=False) ->  # Conv_5
    ConvBlock2D(256, 256) -> Pooling2D(2, 2, "max") ->  # Conv_6

    ConvBlock2D(256, 512, do_pooling=False) ->  # Conv_7
    ConvBlock2D(512, 512) -> Pooling2D(2, 2, "max") ->  # Conv_8

    ConvBlock2D(512, 512, do_pooling=False) ->  # Conv_9
    ConvBlock2D(512, 512) -> Pooling2D(2, 2, "max") ->  # Conv_10
    ```
    Fully Connected Layers:
    ```py
    Flatten ->
    DenseBlock(512 * 7 * 7, 4096) -> DenseBlock(4096, 4096) ->
    Dense(4096, 1000)
    ```
    Output:
    ```py
    Matrix[..., 1000]
    ```
    Parameter Size:
    ```txt
    133,035,712 weights, 12,136 biases -> 133,047,848 params
    ```
    Parameters
    ----------
    `activation` : FuncType, default=Activation.ReLU
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss, default=CrossEntropy()
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer
    `out_features` : int, default=1000
        Number of output features
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.01
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    References
    ----------
    1. Simonyan, Karen, and Andrew Zisserman. "Very Deep Convolutional Networks for
    Large-Scale Image Recognition." arXiv preprint arXiv:1409.1556, 2014.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.ReLU,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 1000,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False,
    ) -> None:
        super().__init__(
            optimizer,
            activation,
            loss,
            initializer,
            out_features,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            dropout_rate,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class VGGNet_16(_vgg._VGGNet_16):
    """
    VGG16 is ont of the variants of the VGG network architecture that was designed
    to enhance image recognition performance through deeper networks with smaller
    convolutional filters. This model was introduced by Karen Simonyan and Andrew
    Zisserman in their paper and is notable for its simplicity and effectiveness
    in image classification tasks.

    Structure
    ---------
    Input:
    ```py
    Tensor[..., 3, 224, 224]
    ```
    Convolutional Blocks:
    ```py
    ConvBlock2D(3, 64, do_pooling=False) ->  # Conv_1
    ConvBlock2D(64, 64) -> Pooling2D(2, 2, "max") ->  # Conv_2

    ConvBlock2D(64, 128, do_pooling=False) ->  # Conv_3
    ConvBlock2D(128, 128) -> Pooling2D(2, 2, "max") ->  # Conv_4

    ConvBlock2D(128, 256, do_pooling=False) ->  # Conv_5
    ConvBlock2D(256, 256, do_pooling=False) ->  # Conv_6
    ConvBlock2D(256, 256) -> Pooling2D(2, 2, "max") ->  # Conv_7

    ConvBlock2D(256, 512, do_pooling=False) ->  # Conv_8
    ConvBlock2D(512, 512, do_pooling=False) ->  # Conv_9
    ConvBlock2D(512, 512) -> Pooling2D(2, 2, "max") ->  # Conv_10

    ConvBlock2D(512, 512, do_pooling=False) ->  # Conv_11
    ConvBlock2D(512, 512, do_pooling=False) ->  # Conv_12
    ConvBlock2D(512, 512) -> Pooling2D(2, 2, "max") ->  # Conv_13
    ```
    Fully Connected Layers:
    ```py
    Flatten ->
    DenseBlock(512 * 7 * 7, 4096) -> DenseBlock(4096, 4096) ->
    Dense(4096, 1000)
    ```
    Output:
    ```py
    Matrix[..., 1000]
    ```
    Parameter Size:
    ```txt
    138,344,128 weights, 13,416 biases -> 138,357,544 params
    ```
    Parameters
    ----------
    `activation` : FuncType, default=Activation.ReLU
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss, default=CrossEntropy()
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer
    `out_features` : int, default=1000
        Number of output features
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.01
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    References
    ----------
    1. Simonyan, Karen, and Andrew Zisserman. "Very Deep Convolutional Networks for
    Large-Scale Image Recognition." arXiv preprint arXiv:1409.1556, 2014.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.ReLU,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 1000,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False,
    ) -> None:
        super().__init__(
            optimizer,
            activation,
            loss,
            initializer,
            out_features,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            dropout_rate,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class VGGNet_19(_vgg._VGGNet_19):
    """
    VGG19 is ont of the variants of the VGG network architecture that was designed
    to enhance image recognition performance through deeper networks with smaller
    convolutional filters. This model was introduced by Karen Simonyan and Andrew
    Zisserman in their paper and is notable for its simplicity and effectiveness
    in image classification tasks.

    Structure
    ---------
    Input:
    ```py
    Tensor[..., 3, 224, 224]
    ```
    Convolutional Blocks:
    ```py
    ConvBlock2D(3, 64, do_pooling=False) ->  # Conv_1
    ConvBlock2D(64, 64) -> Pooling2D(2, 2, "max") ->  # Conv_2

    ConvBlock2D(64, 128, do_pooling=False) ->  # Conv_3
    ConvBlock2D(128, 128) -> Pooling2D(2, 2, "max") ->  # Conv_4

    ConvBlock2D(128, 256, do_pooling=False) ->  # Conv_5
    ConvBlock2D(256, 256, do_pooling=False) ->  # Conv_6
    ConvBlock2D(256, 256, do_pooling=False) ->  # Conv_7
    ConvBlock2D(256, 256) -> Pooling2D(2, 2, "max") ->  # Conv_8

    ConvBlock2D(256, 512, do_pooling=False) ->  # Conv_9
    ConvBlock2D(512, 512, do_pooling=False) ->  # Conv_10
    ConvBlock2D(512, 512, do_pooling=False) ->  # Conv_11
    ConvBlock2D(512, 512) -> Pooling2D(2, 2, "max") ->  # Conv_12

    ConvBlock2D(512, 512, do_pooling=False) ->  # Conv_13
    ConvBlock2D(512, 512, do_pooling=False) ->  # Conv_14
    ConvBlock2D(512, 512, do_pooling=False) ->  # Conv_15
    ConvBlock2D(512, 512) -> Pooling2D(2, 2, "max") ->  # Conv_16
    ```
    Fully Connected Layers:
    ```py
    Flatten ->
    DenseBlock(512 * 7 * 7, 4096) -> DenseBlock(4096, 4096) ->
    Dense(4096, 1000)
    ```
    Output:
    ```py
    Matrix[..., 1000]
    ```
    Parameter Size:
    ```txt
    143,652,544 weights, 14,696 biases -> 143,667,240 params
    ```
    Parameters
    ----------
    `activation` : FuncType, default=Activation.ReLU
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss, default=CrossEntropy()
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer
    `out_features` : int, default=1000
        Number of output features
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.01
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    References
    ----------
    1. Simonyan, Karen, and Andrew Zisserman. "Very Deep Convolutional Networks for
    Large-Scale Image Recognition." arXiv preprint arXiv:1409.1556, 2014.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.ReLU,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 1000,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False,
    ) -> None:
        super().__init__(
            optimizer,
            activation,
            loss,
            initializer,
            out_features,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            dropout_rate,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class Inception_V1(_inception._Inception_V1):
    """
    Inception v1, also known as GoogLeNet, is a deep convolutional neural network
    architecture designed for image classification. It introduces an "Inception
    module," which uses multiple convolutional filters of different sizes in
    parallel to capture various features at different scales. This architecture
    reduces computational costs by using 1x1 convolutions to decrease the number
    of input channels. Inception v1 achieved state-of-the-art results on the
    ImageNet Large Scale Visual Recognition Challenge (ILSVRC) 2014.

    Structure
    ---------
    Input:
    ```py
    Tensor[..., 3, 224, 224]
    ```
    Introductory Convolutions:
    ```py
    ConvBlock2D(3, 64, filter_size=7, pool_mode="max") ->
    ConvBlock2D(64, 64, filter_size=1, do_pooling=False) ->
    ConvBlock2D(64, 192, filter_size=3, pool_mode="max") ->
    ```
    Inception Blocks:
    ```py
    InceptionBlock(192) ->  # Inception_3a
    InceptionBlock(256) ->  # Inception_3b
    Pooling2D(3, 2, mode="max") ->

    InceptionBlock(480) ->  # Inception_4a
    InceptionBlock(512) ->  # Inception_4b
    InceptionBlock(512) ->  # Inception_4c
    InceptionBlock(512) ->  # Inception_4d
    InceptionBlock(528) ->  # Inception_4e
    Pooling2D(3, 2, mode="max") ->

    InceptionBlock(832) ->  # Inception_5a
    InceptionBlock(832) ->  # Inception_5b
    GlobalAvgPooling2D() ->
    ```
    Fully Connected Layers:
    ```py
    Flatten -> Dense(1024, 1000)
    ```
    Output:
    ```py
    Matrix[..., 1000]
    ```
    Parameter Size:
    ```txt
    6,990,272 weights, 8,280 biases -> 6,998,552 params
    ```
    Parameters
    ----------
    `activation` : FuncType, default=Activation.ReLU
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss, default=CrossEntropy()
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer
    `out_features` : int, default=1000
        Number of output features
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.01
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    References
    ----------
    1. Szegedy, Christian, et al. “Going Deeper with Convolutions.” Proceedings
    of the IEEE Conference on Computer Vision and Pattern Recognition (CVPR),
    2015, pp. 1-9.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.ReLU,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 1000,
        batch_size: int = 128,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0.0,
        dropout_rate: float = 0.4,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False,
    ) -> None:
        super().__init__(
            optimizer,
            activation,
            loss,
            initializer,
            out_features,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            dropout_rate,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class Inception_V2(_inception._Inception_V2):
    """
    Inception v2, an improvement of the original Inception architecture,
    enhances computational efficiency and accuracy in deep learning models.
    It introduces the factorization of convolutions and additional
    normalization techniques to reduce the number of parameters and improve
    training stability. These modifications allow for deeper and more
    complex neural networks with improved performance.

    Structure
    ---------
    Input:
    ```py
    Tensor[..., 3, 299, 299]
    ```
    Introductory Convolutions:
    ```py
    ConvBlock2D(3, 32, filter_size=3, stride=2) ->
    ConvBlock2D(32, 32, filter_size=3, stride=1) ->
    ConvBlock2D(32, 64, filter_size=3, stride=1) ->
    Pooling2D(3, 2, mode="max") ->

    ConvBlock2D(64, 80, filter_size=3, stride=1) ->
    ConvBlock2D(80, 192, filter_size=3, stride=2) ->
    ConvBlock2D(192, 288, filter_size=3, stride=1) ->
    ```
    Inception Blocks:
    ```py
    3x InceptionBlockV2A(288) ->  # Inception_3
    InceptionBlockV2R(288) ->  # Inception_Rx1

    5x InceptionBlockV2B(768) ->  # Inception_4
    InceptionBlockV2R(768) ->  # Inception_Rx2

    2x InceptionBlockV2C([1280, 2048]) ->  # Inception_5
    GlobalAvgPooling2D() ->
    ```
    Fully Connected Layers:
    ```py
    Flatten -> Dense(2048, 1000)
    ```
    Output:
    ```py
    Matrix[..., 1000]
    ```
    Parameter Size:
    ```txt
    24,974,688 weights, 20,136 biases -> 24,994,824 params
    ```
    Parameters
    ----------
    `activation` : FuncType, default=Activation.ReLU
        Type of activation function
    `optimizer` : Optimizer
        Type of optimizer for weight update
    `loss` : Loss, default=CrossEntropy()
        Type of loss function
    `initializer` : InitStr, default=None
        Type of weight initializer
    `out_features` : int, default=1000
        Number of output features
    `batch_size` : int, default=100
        Size of a single mini-batch
    `n_epochs` : int, default=100
        Number of epochs for training
    `learning_rate` : float, default=0.01
        Step size during optimization process
    `valid_size` : float, default=0.1
        Fractional size of validation set
    `lambda_` : float, default=0.0
        L2 regularization strength
    `early_stopping` : bool, default=False
        Whether to early-stop the training when the valid score stagnates
    `patience` : int, default=10
        Number of epochs to wait until early-stopping
    `shuffle` : bool, default=True
        Whethter to shuffle the data at the beginning of every epoch

    References
    ----------
    1. Szegedy, Christian, et al. “Going Deeper with Convolutions.” Proceedings
    of the IEEE Conference on Computer Vision and Pattern Recognition (CVPR),
    2015, pp. 1-9.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.ReLU,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 1000,
        batch_size: int = 128,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0.0,
        dropout_rate: float = 0.4,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int | None = None,
        deep_verbose: bool = False,
    ) -> None:
        super().__init__(
            optimizer,
            activation,
            loss,
            initializer,
            out_features,
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            lambda_,
            dropout_rate,
            early_stopping,
            patience,
            shuffle,
            random_state,
            deep_verbose,
        )


class Inception_V3(_inception._Inception_V3):
    NotImplemented



from luma.preprocessing.encoder import OneHotEncoder
from luma.neural.optimizer import AdamOptimizer
from luma.neural.network import *

import numpy as np

np.random.seed(42)

model = VGGNet_19(
    optimizer=AdamOptimizer(),
)
model.summarize(in_shape=(-1, 3, 224, 224))

# X = np.random.randn(100, 3, 224, 224)
# y = np.array(list(range(10)) * 10)
# y = OneHotEncoder().fit_transform(y.reshape(-1, 1))

# out = model.fit(X, y)



from typing import Iterator, Tuple
import numpy as np

from luma.interface.typing import Matrix, Vector, TensorLike


__all__ = ("TrainTestSplit", "BatchGenerator")


class TrainTestSplit:
    """
    Splits the original dataset into the train set and the test set.

    Parameters
    ----------
    `X` : Matrix
        Feature data
    `y` : Vector
        Target data
    `test_size` : int or float, default=0.3
        Proportional size of the test set
    `shuffle` : bool, default=True
        Whether to shuffle the dataset
    `stratify` : bool, default=False
        Whether to perform stratified split
    `random_state` : int, optional, default=None
        Seed for random sampling for split

    Properties
    ----------
    `get` : tuple[Matrix, Matrix, Vector, Vector]
        Returns the split data as a 4-tuple

    Notes
    -----
    For `stratify == True`, one-hot encoded `y` is not compatible.

    Examples
    --------
    ```py
    X_train, X_test, y_train, y_test = TrainTestSplit(X, y, ...).get
    ```
    """

    def __init__(
        self,
        X: Matrix,
        y: Vector,
        test_size: int | float = 0.3,
        shuffle: bool = True,
        stratify: bool = False,
        random_state: int = None,
    ) -> None:
        self.X = X
        self.y = y
        self.test_size = test_size
        self.shuffle = shuffle
        self.stratify = stratify
        self.random_state = random_state

    @property
    def get(self) -> Tuple[Matrix, Matrix, Vector, Vector]:
        if self.X.shape[0] != self.y.shape[0]:
            raise ValueError("Sample size mismatch between 'X' and 'y'!")

        if self.stratify:
            return self._stratified_split()
        else:
            return self._split()

    def _split(self) -> Tuple[Matrix, Matrix, Vector, Vector]:
        n_samples = self.X.shape[0]
        indices = np.arange(n_samples)

        if isinstance(self.test_size, float):
            num_test_samples = int(self.test_size * n_samples)
        else:
            num_test_samples = self.test_size

        if self.random_state is not None:
            np.random.seed(self.random_state)
        if self.shuffle:
            np.random.shuffle(indices)

        test_indices = indices[:num_test_samples]
        train_indices = indices[num_test_samples:]

        X_train = self.X[train_indices]
        X_test = self.X[test_indices]
        y_train = self.y[train_indices]
        y_test = self.y[test_indices]

        return X_train, X_test, y_train, y_test

    def _stratified_split(self) -> Tuple[Matrix, Matrix, Vector, Vector]:
        if isinstance(self.test_size, float):
            test_size_func = lambda u: int(self.test_size * len(u))
        else:
            total_size = len(self.y)
            proportion = self.test_size / total_size
            test_size_func = lambda u: int(proportion * len(u))

        unique_classes, y_indices = np.unique(self.y, return_inverse=True)
        train_indices, test_indices = [], []

        for class_index in unique_classes:
            class_mask = y_indices == class_index
            class_indices = np.where(class_mask)[0]

            if self.shuffle:
                np.random.shuffle(class_indices)

            num_test_samples = test_size_func(class_indices)
            test_indices.extend(class_indices[:num_test_samples])
            train_indices.extend(class_indices[num_test_samples:])

        if self.shuffle:
            np.random.shuffle(train_indices)
            np.random.shuffle(test_indices)

        X_train = self.X[train_indices]
        X_test = self.X[test_indices]
        y_train = self.y[train_indices]
        y_test = self.y[test_indices]

        return X_train, X_test, y_train, y_test


class BatchGenerator:
    """
    A class for generating mini-batches of data for training machine
    learning models including neural networks.

    Parameters
    ----------
    `X` : TensorLike
        Input features
    `y` : TensorLike
        Targets or labels
    `batch_size` : int, default=100
        Size of a mini-batch
    `shuffle` : bool, default=True
        Whether to shuffle the data for every batch generation

    Iterator
    --------
    The iterator of `BatchGenerator` returns a 2-tuple of
    `TensorLike` objects.
    ```py
    def __iter__(self) -> Iterator[Tuple[TensorLike, TensorLike]]
    ```
    Examples
    --------
    An instance of `BatchGenerator` can be used as an iterator.

    - With instantiation:

        ```py
        batch_gen = BatchGenerator(X, y, batch_size=100)
        for X_batch, y_batch in batch_gen:
            pass
        ```
    - Without instantiation:

        ```py
        for X_batch, y_batch in BatchGenerator(X, y, batch_size=100):
            pass
        ```
    """

    def __init__(
        self,
        X: TensorLike,
        y: TensorLike,
        batch_size: int = 100,
        shuffle: bool = True,
    ) -> None:
        self.X = X
        self.y = y
        self.batch_size = batch_size
        self.shuffle = shuffle

        self.n_samples = X.shape[0]
        self.n_batches = self.n_samples // batch_size

        if self.n_samples % batch_size != 0:
            self.n_batches += 1

        self.indices = np.arange(self.n_samples)
        if self.shuffle:
            np.random.shuffle(self.indices)

    def __iter__(self) -> Iterator[Tuple[TensorLike, TensorLike]]:
        for i in range(self.n_batches):
            start_idx = i * self.batch_size
            end_idx = min((i + 1) * self.batch_size, self.n_samples)

            batch_indices = self.indices[start_idx:end_idx]
            X_batch = self.X[batch_indices]
            y_batch = self.y[batch_indices]

            yield X_batch, y_batch



from typing import Any


__all__ = (
    "NotFittedError",
    "UnsupportedParameterError",
    "NotConvergedError",
    "ModelExtensionError",
    "InvalidRangeError",
)


def get_name(model: Any) -> str:
    return type(model).__name__


class NotFittedError(Exception):
    def __init__(self, model: object | Any) -> None:
        super().__init__(
            f"'{get_name(model)}' is not fitted!"
            + f" Call '{get_name(model)}.fit()' to fit the model."
        )


class UnsupportedParameterError(Exception):
    def __init__(self, param: Any) -> None:
        super().__init__(f"'{param}' is unsupported!")


class NotConvergedError(Exception):
    def __init__(self, model: object | Any) -> None:
        super().__init__(
            f"'{get_name(model)}' did not converged!"
            + f" Try setting '{get_name(model)}.tol' to bigger value."
        )


class ModelExtensionError(Exception):
    def __init__(self, filename: str) -> None:
        super().__init__(f"'{filename}' is not a '.luma' model file!")


class InvalidRangeError(Exception):
    def __init__(self, param: Any, param_name: str, param_range: str) -> None:
        super().__init__(
            f"'{param_name}' of {param} should "
            + f"fall within the range of ({param_range})"
        )
